# -*- coding: utf-8 -*-
# This file is part of Eigen, a lightweight C++ template library
# for linear algebra.
#
# Copyright (C) 2021 Huang, Zhaoquan <zhaoquan2008@hotmail.com>
# Copyright (C) 2025 Prince Mathew, Joseph <josephrpincemathew@outlook.com
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

# Pretty printers for Eigen::Matrix to use with LLDB debugger
#
# Usage:
# 1. Add the following line (change it according to the path to this file)
#    to the file ~/.lldbinit (create one if it doesn't exist):
#        `command script import /path/to/eigenlldb.py`
# 2. Inspect the variables in LLDB command line
#        `frame variable`

import lldb
from typing import List
import bisect


def __lldb_init_module(debugger, internal_dict):
    debugger.HandleCommand(
        "type synthetic add -x ^Eigen::Matrix<.*> --python-class eigenlldb.EigenMatrixChildProvider")
    debugger.HandleCommand(
        "type summary add -e -x ^Eigen::Matrix<.*> --python-function eigenlldb.eigen_matrix_summary_provider")
    debugger.HandleCommand(
        "type synthetic add -x ^Eigen::SparseMatrix<.*> --python-class eigenlldb.EigenSparseMatrixChildProvider")

class EigenMatrixProvider:
    def __init__(self, valobj: lldb.SBValue, internal_dict):

        if valobj.IsSynthetic():
            self._valobj: lldb.SBValue = valobj.GetNonSyntheticValue()
        else:
            self._valobj: lldb.SBValue = valobj

        _valtype = self._valobj.GetType().GetCanonicalType()

        scalar_type = _valtype.GetTemplateArgumentType(0)
        if not scalar_type.IsValid():
            # In the case that scalar_type is invalid on LLDB 9.0 on Windows with CLion
            storage = self._valobj.GetChildMemberWithName("m_storage")
            data = storage.GetChildMemberWithName("m_data")
            data_type = data.GetType()
            if data_type.IsPointerType():
                scalar_type = data.GetType().GetPointeeType()
            else:
                scalar_type = data.GetChildMemberWithName("array").GetType().GetArrayElementType()
        self._scalar_type: lldb.SBType = scalar_type
        self._scalar_size: int = self._scalar_type.GetByteSize()

        name = _valtype.GetName()
        template_begin = name.find("<")
        template_end = name.rfind(">")
        template_args = name[(template_begin + 1):template_end].split(",")

        template_args = [arg.strip() for arg in template_args]

        # Provide default values for missing template arguments
        # Eigen::Matrix<Scalar, Rows, Cols, Options, MaxRows, MaxCols>
        try:
            self._rows_compile_time: int = int(template_args[1]) if len(template_args) > 1 else -1
        except (ValueError, IndexError):
            self._rows_compile_time: int = -1

        try:
            self._cols_compile_time: int = int(template_args[2]) if len(template_args) > 2 else -1
        except (ValueError, IndexError):
            self._cols_compile_time: int = -1

        try:
            self._options: int = int(template_args[3]) if len(template_args) > 3 else 0
            self._row_major: bool = (self._options & 1) != 0
        except (ValueError, IndexError):
            self._options = 0
            self._row_major: bool = False

        try:
            max_rows = int(template_args[4]) if len(template_args) > 4 else self._rows_compile_time
        except (ValueError, IndexError):
            max_rows = self._rows_compile_time

        try:
            max_cols = int(template_args[5]) if len(template_args) > 5 else self._cols_compile_time
        except (ValueError, IndexError):
            max_cols = self._cols_compile_time

        self._fixed_storage: bool = (max_rows != -1 and max_cols != -1)

    def _cols(self):
        if self._cols_compile_time == -1:
            storage = self._valobj.GetChildMemberWithName("m_storage")
            cols = storage.GetChildMemberWithName("m_cols").GetValueAsUnsigned()
            return cols
        else:
            return self._cols_compile_time

    def _rows(self):
        if self._rows_compile_time == -1:
            storage = self._valobj.GetChildMemberWithName("m_storage")
            rows = storage.GetChildMemberWithName("m_rows").GetValueAsUnsigned()
            return rows
        else:
            return self._rows_compile_time
    def get_summary(self):
        if self._rows() == 0 or self._cols() == 0:
            return f"size=[0x0, order={'row_major' if self._row_major else 'column_major'}]"
        else:
            return f"size=[{self._rows()}x{self._cols()}, order={'row_major' if self._row_major else 'column_major'}] "


def eigen_matrix_summary_provider(valobj, dict):
    prov = EigenMatrixProvider(valobj, dict)
    return prov.get_summary()

class EigenMatrixChildProvider(EigenMatrixProvider):

    def num_children(self):
        return self._cols() * self._rows()

    def get_child_index(self, name):
        pass

    def get_child_at_index(self, index):
        storage = self._valobj.GetChildMemberWithName("m_storage")
        data = storage.GetChildMemberWithName("m_data")
        offset = self._scalar_size * index

        if self._row_major:
            row = index // self._cols()
            col = index % self._cols()
        else:
            row = index % self._rows()
            col = index // self._rows()
        if self._fixed_storage:
            data = data.GetChildMemberWithName("array")
        if self._cols() == 1:
            name = '[{}]'.format(row)
        elif self._rows() == 1:
            name = '[{}]'.format(col)
        else:
            name = '[{},{}]'.format(row, col)
        return data.CreateChildAtOffset(
            name, offset, self._scalar_type
        )

class EigenSparseMatrixChildProvider:
    _valobj: lldb.SBValue
    _scalar_type: lldb.SBType
    _scalar_size: int
    _index_type: lldb.SBType
    _index_size: int
    _row_major: bool

    _outer_size: int
    _nnz: int
    _values: lldb.SBValue
    _inner_indices: lldb.SBValue
    _outer_starts: lldb.SBValue
    _inner_nnzs: lldb.SBValue
    _compressed: bool

    # Index of the first synthetic child under each outer index
    _child_indices: List[int]

    def __init__(self, valobj, internal_dict):
        self._valobj = valobj
        valtype = valobj.GetType().GetCanonicalType()
        scalar_type = valtype.GetTemplateArgumentType(0)
        if not scalar_type.IsValid():
            # In the case that scalar_type is invalid on LLDB 9.0 on Windows with CLion
            data = valobj.GetChildMemberWithName("m_data")
            values = data.GetChildMemberWithName("m_values")
            scalar_type = values.GetType().GetPointeeType()
        self._scalar_type = scalar_type
        self._scalar_size = scalar_type.GetByteSize()

        index_type = valtype.GetTemplateArgumentType(2)
        if not index_type.IsValid():
            # In the case that scalar_type is invalid on LLDB 9.0 on Windows with CLion
            outer_starts = valobj.GetChildMemberWithName("m_outerIndex")
            index_type = outer_starts.GetType().GetPointeeType()
        self._index_type = index_type
        self._index_size = index_type.GetByteSize()

        name = valtype.GetName()
        template_begin = name.find("<")
        template_end = name.find(">")
        template_args = name[(template_begin + 1):template_end].split(",")
        self._row_major = (int(template_args[1]) & 1) != 0

    def num_children(self):
        return self._nnz + 2

    def get_child_index(self, name):
        pass

    def get_child_at_index(self, index):
        if index == 0:
            name = "rows" if self._row_major else "cols"
            return self._valobj.GetChildMemberWithName("m_outerSize") \
                .CreateChildAtOffset(name, 0, self._index_type)
        elif index == 1:
            name = "cols" if self._row_major else "rows"
            return self._valobj.GetChildMemberWithName("m_innerSize") \
                .CreateChildAtOffset(name, 0, self._index_type)
        else:
            index = index - 2
        outer_index = bisect.bisect_right(self._child_indices, index) - 1
        total_nnzs = self._child_indices[outer_index]
        if self._compressed:
            item_index = index
            inner_index = self._inner_indices \
                .CreateChildAtOffset("", item_index * self._index_size, self._index_type) \
                .GetValueAsUnsigned()
            return self._values \
                .CreateChildAtOffset(self._child_name(outer_index, inner_index),
                                     item_index * self._scalar_size,
                                     self._scalar_type)
        else:
            index_begin = self._outer_starts \
                .CreateChildAtOffset("", outer_index * self._index_size, self._index_type) \
                .GetValueAsUnsigned()
            item_index = index - total_nnzs + index_begin
            inner_index = self._inner_indices \
                .CreateChildAtOffset("", item_index * self._index_size, self._index_type) \
                .GetValueAsUnsigned()
            return self._values \
                .CreateChildAtOffset(self._child_name(outer_index, inner_index),
                                     item_index * self._scalar_size,
                                     self._scalar_type)

    def update(self):
        valobj = self._valobj
        self._outer_size = valobj.GetChildMemberWithName("m_outerSize").GetValueAsUnsigned()
        data = valobj.GetChildMemberWithName("m_data")
        self._values = data.GetChildMemberWithName("m_values")
        self._inner_indices = data.GetChildMemberWithName("m_indices")
        self._outer_starts = valobj.GetChildMemberWithName("m_outerIndex")
        self._inner_nnzs = valobj.GetChildMemberWithName("m_innerNonZeros")

        self._compressed = self._inner_nnzs.GetValueAsUnsigned() == 0

        total_nnzs = 0
        child_indices = [0]
        for outer_index in range(self._outer_size):
            if self._compressed:
                index_end = self._outer_starts \
                    .CreateChildAtOffset("", (outer_index + 1) * self._index_size, self._index_type) \
                    .GetValueAsUnsigned()
                total_nnzs = index_end
                child_indices.append(total_nnzs)
            else:
                nnzs = self._inner_nnzs \
                    .CreateChildAtOffset("", outer_index * self._index_size, self._index_type) \
                    .GetValueAsUnsigned()
                total_nnzs = total_nnzs + nnzs
                child_indices.append(total_nnzs)
        self._child_indices = child_indices
        self._nnz = total_nnzs

    def _child_name(self, outer_index, inner_index):
        if self._row_major:
            return "[{0},{1}]".format(outer_index, inner_index)
        else:
            return "[{1},{0}]".format(outer_index, inner_index)



__all__ = ['Normality', 'Bootstrap', 'PERMBU', 'Conformal']


import warnings
from enum import Enum

import numpy as np
import scipy.sparse as sp
from scipy.stats import norm
from sklearn.preprocessing import OneHotEncoder

from .utils import (
    _ma_cov,
    _shrunk_covariance_schaferstrimmer_no_nans,
    _shrunk_covariance_schaferstrimmer_with_nans,
    is_strictly_hierarchical,
)


class CovarianceType(str, Enum):
    """Covariance estimation method for Normality probabilistic reconciliation.

    Attributes:
        DIAGONAL: Uses the W matrix diagonal with correlation scaling.
            This is the default and backward-compatible option.
            The W matrix is required.
        FULL: Uses full empirical covariance computed from residuals.
            Requires the `residuals` parameter. The W matrix is ignored.
            Warning: May produce non-positive-definite matrices when
            n_series > n_observations.
        SHRINK: Uses Schäfer-Strimmer shrinkage estimator for better numerical
            stability. Requires the `residuals` parameter. The W matrix is ignored.
            Recommended when n_series is large relative to n_observations.
    """

    DIAGONAL = "diagonal"
    FULL = "full"
    SHRINK = "shrink"


class Normality:
    r"""Normality Probabilistic Reconciliation Class.

    The Normality method leverages the Gaussian Distribution linearity, to
    generate hierarchically coherent prediction distributions. This class is
    meant to be used as the `sampler` input as other `HierarchicalForecast` [reconciliation classes](./methods).

    Given base forecasts under a normal distribution:
    ```math
    \hat{y}_{h} \sim \mathrm{N}(\hat{\boldsymbol{\mu}}, \hat{\mathbf{W}}_{h})
    ```

    The reconciled forecasts are also normally distributed:

    ```math
    \tilde{y}_{h} \sim \mathrm{N}(\mathbf{S}\mathbf{P}\hat{\boldsymbol{\mu}},
    \mathbf{S}\mathbf{P}\hat{\mathbf{W}}_{h} \mathbf{P}^{\intercal} \mathbf{S}^{\intercal})
    ```

    Args:
        S (Union[np.ndarray, sp.spmatrix]): Summing matrix of size (`base`, `bottom`).
        P (Union[np.ndarray, sp.spmatrix]): Reconciliation matrix of size (`bottom`, `base`).
        y_hat (np.ndarray): Point forecasts values of size (`base`, `horizon`).
        sigmah (np.ndarray): Forecast standard dev. of size (`base`, `horizon`).
        W (Union[np.ndarray, sp.spmatrix], optional): Hierarchical covariance matrix of size
            (`base`, `base`). Required when `covariance_type='diagonal'` (default).
            **Ignored** when `covariance_type` is `'full'` or `'shrink'` (covariance is
            computed from residuals instead). Default is None.
        seed (int, optional): Random seed for numpy generator's replicability. Default is 0.
        covariance_type (Union[str, CovarianceType], optional): Type of covariance estimator.
            Can be a string or CovarianceType enum. Options are:

            - `'diagonal'` / `CovarianceType.DIAGONAL`: Uses the W matrix diagonal with
              correlation scaling (default, backward compatible). W is required.
            - `'full'` / `CovarianceType.FULL`: Uses full empirical covariance from residuals.
              W is ignored. Warning: may be non-positive-definite if n_series > n_observations.
            - `'shrink'` / `CovarianceType.SHRINK`: Uses Schäfer-Strimmer shrinkage estimator.
              W is ignored. Recommended for numerical stability with many series.

            Default is `'diagonal'`.
        residuals (np.ndarray, optional): Insample residuals of size (`base`, `obs`).
            Required when `covariance_type` is `'full'` or `'shrink'`. Default is None.
        shrinkage_ridge (float, optional): Ridge parameter for shrinkage covariance estimator.
            Only used when `covariance_type='shrink'`. A warning is issued if provided
            with other covariance types. Default is 2e-8.

    Raises:
        ValueError: If `covariance_type` is invalid.
        ValueError: If `covariance_type='diagonal'` and `W` is None.
        ValueError: If `covariance_type` is `'full'` or `'shrink'` and `residuals` is None.
        ValueError: If `residuals` shape doesn't match expected (`base`, `obs`).
        ValueError: If `residuals` has fewer than 2 observations.
        ValueError: If `residuals` is empty.
        ValueError: If any series in `residuals` has all NaN values.

    Warnings:
        UserWarning: If `shrinkage_ridge` is provided but `covariance_type` is not `'shrink'`.
        UserWarning: If `W` is provided but `covariance_type` is not `'diagonal'` (W is ignored).
        UserWarning: If any series has zero or near-zero variance (may affect correlation estimates).
        UserWarning: If `covariance_type='full'` and n_series > n_observations (non-PSD risk).

    References:
        - [Panagiotelis A., Gamakumara P. Athanasopoulos G., and Hyndman R. J. (2022).
          "Probabilistic forecast reconciliation: Properties, evaluation and score optimisation".
          European Journal of Operational Research.](https://www.sciencedirect.com/science/article/pii/S0377221722006087)
        - [Schäfer, Juliane, and Korbinian Strimmer. "A Shrinkage Approach to Large-Scale
          Covariance Matrix Estimation". Statistical Applications in Genetics and Molecular
          Biology 4, no. 1 (2005).](https://doi.org/10.2202/1544-6115.1175)

    Examples:
        >>> # Using diagonal covariance (default, backward compatible)
        >>> normality = Normality(S=S, P=P, y_hat=y_hat, sigmah=sigmah, W=W)
        >>> samples = normality.get_samples(num_samples=100)

        >>> # Using full empirical covariance from residuals
        >>> normality = Normality(
        ...     S=S, P=P, y_hat=y_hat, sigmah=sigmah,
        ...     covariance_type="full", residuals=residuals
        ... )

        >>> # Using shrinkage covariance (recommended for stability)
        >>> normality = Normality(
        ...     S=S, P=P, y_hat=y_hat, sigmah=sigmah,
        ...     covariance_type=CovarianceType.SHRINK, residuals=residuals
        ... )
    """

    # Numerical stability constants
    # Minimum standard deviation threshold to avoid division by zero when
    # converting covariance to correlation. Values below this are clamped.
    # Chosen as sqrt(machine epsilon) for float64 ≈ 1.5e-8, rounded to 1e-8.
    _MIN_STD_THRESHOLD: float = 1e-8

    # Default ridge parameter for shrinkage covariance estimator.
    # Provides numerical stability by ensuring diagonal elements are bounded away from zero.
    # Based on Schäfer-Strimmer (2005) recommendations for genomic data.
    _DEFAULT_SHRINKAGE_RIDGE: float = 2e-8

    def __init__(
        self,
        S: np.ndarray | sp.spmatrix,
        P: np.ndarray | sp.spmatrix,
        y_hat: np.ndarray,
        sigmah: np.ndarray,
        W: np.ndarray | sp.spmatrix | None = None,
        seed: int = 0,
        covariance_type: str | CovarianceType = "diagonal",
        residuals: np.ndarray | None = None,
        shrinkage_ridge: float = _DEFAULT_SHRINKAGE_RIDGE,
    ):
        # Normalize covariance_type to CovarianceType enum
        if isinstance(covariance_type, str):
            covariance_type_lower = covariance_type.lower()
            try:
                covariance_type = CovarianceType(covariance_type_lower)
            except ValueError as e:
                valid_types = [t.value for t in CovarianceType]
                raise ValueError(
                    f"Unknown covariance_type `{covariance_type}`. "
                    f"Choose from {valid_types} or use CovarianceType enum."
                ) from e
        elif not isinstance(covariance_type, CovarianceType):
            raise ValueError(
                f"covariance_type must be a string or CovarianceType enum, "
                f"got {type(covariance_type).__name__}."
            )

        # Get number of series from S matrix
        n_series = S.shape[0]

        # Validate W for diagonal covariance type
        if covariance_type == CovarianceType.DIAGONAL:
            if W is None:
                raise ValueError(
                    "covariance_type='diagonal' requires `W` parameter. "
                    "Provide hierarchical covariance matrix of size (`base`, `base`), "
                    "or use covariance_type='full' or 'shrink' with residuals."
                )
            # Check W has valid diagonal
            w_diag = W.diagonal() if hasattr(W, 'diagonal') else np.diag(W)
            if np.any(w_diag <= 0):
                raise ValueError(
                    "W matrix has non-positive diagonal elements. "
                    "Covariance diagonal must be strictly positive."
                )
            if np.any(np.isnan(w_diag)):
                raise ValueError("W matrix contains NaN values on the diagonal.")

        # Warn if W is provided but will be ignored
        if covariance_type != CovarianceType.DIAGONAL and W is not None:
            warnings.warn(
                f"W parameter is ignored when covariance_type='{covariance_type.value}'. "
                "Covariance will be computed from residuals instead.",
                UserWarning,
                stacklevel=2,
            )

        # Warn if shrinkage_ridge is provided but not used
        if covariance_type != CovarianceType.SHRINK and shrinkage_ridge != self._DEFAULT_SHRINKAGE_RIDGE:
            warnings.warn(
                f"shrinkage_ridge parameter is only used when covariance_type='shrink'. "
                f"Current covariance_type='{covariance_type.value}', shrinkage_ridge will be ignored.",
                UserWarning,
                stacklevel=2,
            )

        # Validate residuals for full/shrink types
        if covariance_type in (CovarianceType.FULL, CovarianceType.SHRINK):
            if residuals is None:
                raise ValueError(
                    f"covariance_type='{covariance_type.value}' requires `residuals` parameter. "
                    "Provide insample residuals of size (`base`, `obs`)."
                )

            # Validate residuals is a 2D array
            if residuals.ndim != 2:
                raise ValueError(
                    f"residuals must be a 2D array of shape (`base`, `obs`), "
                    f"got {residuals.ndim}D array with shape {residuals.shape}."
                )

            # Validate residuals shape matches number of series
            if residuals.shape[0] != n_series:
                raise ValueError(
                    f"residuals shape mismatch: first dimension ({residuals.shape[0]}) "
                    f"must match number of series in S ({n_series}). "
                    f"Expected residuals shape: ({n_series}, obs)."
                )

            # Validate residuals has enough observations
            if residuals.shape[1] == 0:
                raise ValueError(
                    "residuals is empty (0 observations). "
                    "Provide residuals with at least 2 observations."
                )

            if residuals.shape[1] < 2:
                raise ValueError(
                    f"residuals has only {residuals.shape[1]} observation(s). "
                    "At least 2 observations are required to compute covariance."
                )

            # Check for all-NaN series
            nan_mask = np.isnan(residuals)
            nan_counts_per_series = nan_mask.sum(axis=1)
            all_nan_series = nan_counts_per_series == residuals.shape[1]
            if np.any(all_nan_series):
                all_nan_indices = np.where(all_nan_series)[0]
                raise ValueError(
                    f"residuals contains series with all NaN values at indices: "
                    f"{all_nan_indices.tolist()}. Each series must have at least "
                    "2 non-NaN observations."
                )

            # Check for series with insufficient non-NaN observations
            non_nan_counts = residuals.shape[1] - nan_counts_per_series
            insufficient_obs = non_nan_counts < 2
            if np.any(insufficient_obs):
                insufficient_indices = np.where(insufficient_obs)[0]
                raise ValueError(
                    f"residuals contains series with fewer than 2 non-NaN observations "
                    f"at indices: {insufficient_indices.tolist()}. Each series must have "
                    "at least 2 non-NaN observations to compute covariance."
                )

            # Warn about n_series > n_observations for full covariance
            n_obs = residuals.shape[1] - nan_counts_per_series.max()
            if covariance_type == CovarianceType.FULL and n_series > n_obs:
                warnings.warn(
                    f"covariance_type='full' with n_series ({n_series}) > n_observations ({n_obs}) "
                    "may produce a non-positive-definite covariance matrix. "
                    "Consider using covariance_type='shrink' for better numerical stability.",
                    UserWarning,
                    stacklevel=2,
                )

        self.S = S
        self.P = P
        self.y_hat = y_hat
        self.covariance_type = covariance_type
        self.residuals = residuals
        self.shrinkage_ridge = shrinkage_ridge

        if isinstance(P, sp.linalg.LinearOperator) and sp.issparse(S):
            self.SP = sp.linalg.aslinearoperator(self.S) @ self.P
        else:
            self.SP = self.S @ self.P
        self.W = W
        self.sigmah = sigmah
        self.seed = seed

        # Compute correlation matrix based on covariance_type
        R1 = self._compute_correlation_matrix(covariance_type, W, residuals, shrinkage_ridge)

        # Store for potential inspection
        self._correlation_matrix = R1

        # Using elementwise multiplication
        cov_recs = []
        sigmah_recs = []
        for sigma in self.sigmah.T:
            # Broadcast sigma to create a matrix of pairwise products
            sigma_matrix = np.outer(sigma, sigma)
            # Element-wise multiplication with correlation matrix
            if sp.issparse(R1):
                cov_matrix = R1.multiply(sigma_matrix).toarray()
                cov_rec = self.SP @ cov_matrix @ self.SP.T
            else:
                # If R1 is dense, use numpy multiplication
                cov_matrix = R1 * sigma_matrix
                cov_rec = self.SP @ cov_matrix @ self.SP.T
            cov_recs.append(cov_rec)
            diag_cov = cov_rec.diagonal()
            if np.any(diag_cov < 0):
                n_neg = np.count_nonzero(diag_cov < 0)
                warnings.warn(
                    f"Detected {n_neg} negative variance value(s) in reconciled covariance "
                    "matrix diagonal. This indicates a non-positive-definite covariance "
                    "matrix. Negative variances will be clamped to zero.",
                    RuntimeWarning,
                )
            sigmah_recs.append(np.sqrt(np.maximum(diag_cov, 0)))

        self.sigmah_rec = np.hstack(sigmah_recs).reshape(-1, self.sigmah.shape[0]).T
        self.cov_rec = cov_recs

    def _compute_correlation_matrix(
        self,
        covariance_type: CovarianceType,
        W: np.ndarray | sp.spmatrix | None,
        residuals: np.ndarray | None,
        shrinkage_ridge: float,
    ) -> np.ndarray:
        """Compute correlation matrix based on covariance type.

        Args:
            covariance_type: The covariance estimation method to use.
            W: Hierarchical covariance matrix (used for diagonal type).
            residuals: Insample residuals (used for full/shrink types).
            shrinkage_ridge: Ridge parameter for shrinkage estimator.

        Returns:
            Correlation matrix of size (n_series, n_series).
        """
        if covariance_type == CovarianceType.DIAGONAL and W is not None:
            # Original behavior: use W diagonal with correlation scaling
            diag = W.diagonal() if hasattr(W, "diagonal") else np.diag(W)
            std_ = np.sqrt(diag)
            R1 = W / np.outer(std_, std_)
        elif covariance_type == CovarianceType.FULL:
            R1 = self._compute_full_correlation(residuals)
        elif covariance_type == CovarianceType.SHRINK:
            R1 = self._compute_shrink_correlation(residuals, shrinkage_ridge)

        return R1

    def _compute_full_correlation(self, residuals: np.ndarray) -> np.ndarray:
        """Compute full empirical correlation matrix from residuals.

        Args:
            residuals: Insample residuals of size (n_series, n_obs).

        Returns:
            Correlation matrix of size (n_series, n_series).
        """
        nan_mask = np.isnan(residuals)
        if np.any(nan_mask):
            cov_matrix = _ma_cov(residuals, ~nan_mask)
        else:
            cov_matrix = np.cov(residuals)

        # Handle edge case: np.cov returns scalar for single series
        if cov_matrix.ndim == 0:
            cov_matrix = np.array([[cov_matrix]])

        return self._covariance_to_correlation(cov_matrix)

    def _compute_shrink_correlation(
        self, residuals: np.ndarray, shrinkage_ridge: float
    ) -> np.ndarray:
        """Compute shrinkage correlation matrix using Schäfer-Strimmer method.

        Args:
            residuals: Insample residuals of size (n_series, n_obs).
            shrinkage_ridge: Ridge parameter for numerical stability.

        Returns:
            Correlation matrix of size (n_series, n_series).
        """
        nan_mask = np.isnan(residuals)
        if np.any(nan_mask):
            cov_matrix = _shrunk_covariance_schaferstrimmer_with_nans(
                residuals, ~nan_mask, shrinkage_ridge
            )
        else:
            cov_matrix = _shrunk_covariance_schaferstrimmer_no_nans(
                residuals, shrinkage_ridge
            )

        return self._covariance_to_correlation(cov_matrix)

    def _covariance_to_correlation(self, cov_matrix: np.ndarray) -> np.ndarray:
        """Convert covariance matrix to correlation matrix with numerical safeguards.

        Args:
            cov_matrix: Covariance matrix of size (n_series, n_series).

        Returns:
            Correlation matrix of size (n_series, n_series).

        Warns:
            UserWarning: If any series has zero or near-zero variance.
        """
        std_ = np.sqrt(np.diag(cov_matrix))

        # Check for zero/near-zero variance series
        low_variance_mask = std_ < self._MIN_STD_THRESHOLD
        if np.any(low_variance_mask):
            low_variance_indices = np.where(low_variance_mask)[0]
            warnings.warn(
                f"Series at indices {low_variance_indices.tolist()} have zero or "
                f"near-zero variance (std < {self._MIN_STD_THRESHOLD}). "
                "Correlation estimates for these series may be unreliable. "
                "Consider removing constant series or using regularization.",
                UserWarning,
                stacklevel=4,
            )
            # Clamp to minimum threshold to avoid division by zero
            std_[low_variance_mask] = self._MIN_STD_THRESHOLD

        # Convert to correlation
        correlation = cov_matrix / np.outer(std_, std_)

        # Ensure diagonal is exactly 1.0 (numerical precision fix)
        np.fill_diagonal(correlation, 1.0)

        # Clamp off-diagonal to [-1, 1] for numerical stability
        correlation = np.clip(correlation, -1.0, 1.0)

        return correlation

    def get_samples(self, num_samples: int):
        """Normality Coherent Samples.

        Obtains coherent samples under the Normality assumptions.

        Args:
            num_samples (int): number of samples generated from coherent distribution.

        Returns:
            samples (np.ndarray): Coherent samples of size (`base`, `horizon`, `num_samples`).
        """
        rng = np.random.default_rng(self.seed)
        n_series, n_horizon = self.y_hat.shape
        samples = np.empty(shape=(num_samples, n_series, n_horizon))
        for t in range(n_horizon):
            with warnings.catch_warnings():
                # Avoid 'RuntimeWarning: covariance is not positive-semidefinite.'
                # By definition the multivariate distribution is not full-rank
                partial_samples = rng.multivariate_normal(
                    mean=self.SP @ self.y_hat[:, t],
                    cov=self.cov_rec[t],
                    size=num_samples,
                )
            samples[:, :, t] = partial_samples

        # [samples, N, H] -> [N, H, samples]
        samples = samples.transpose((1, 2, 0))
        return samples

    def get_prediction_levels(self, res, level):
        """Adds reconciled forecast levels to results dictionary

        Args:
            res (dict): Results dictionary to update.
            level (list): Confidence levels.

        Returns:
            dict: Updated results dictionary.
        """
        res["sigmah"] = self.sigmah_rec
        level = np.asarray(level)
        z = norm.ppf(0.5 + level / 200)
        for zs, lv in zip(z, level, strict=False):
            res[f"lo-{lv}"] = res["mean"] - zs * self.sigmah_rec
            res[f"hi-{lv}"] = res["mean"] + zs * self.sigmah_rec
        return res

    def get_prediction_quantiles(self, res, quantiles):
        """Adds reconciled forecast quantiles to results dictionary

        Args:
            res (dict): Results dictionary to update.
            quantiles (np.ndarray): Quantiles to compute.

        Returns:
            dict: Updated results dictionary.
        """
        # [N,H,None] + [None None,Q] * [N,H,None] -> [N,H,Q]
        z = norm.ppf(quantiles)
        res["sigmah"] = self.sigmah_rec
        res["quantiles"] = (
            res["mean"][:, :, None] + z[None, None, :] * self.sigmah_rec[:, :, None]
        )
        return res


class Bootstrap:
    r"""Bootstrap Probabilistic Reconciliation Class.

    This method goes beyond the normality assumption for the base forecasts,
    the technique simulates future sample paths and uses them to generate
    base sample paths that are latered reconciled. This clever idea and its
    simplicity allows to generate coherent bootstraped prediction intervals
    for any reconciliation strategy. This class is meant to be used as the `sampler`
    input as other `HierarchicalForecast` [reconciliation classes](./methods).

    Given a boostraped set of simulated sample paths:
    ```math
    \hat{\mathbf{y}}^{[1]}_{\\tau}, \dots ,\hat{\mathbf{y}}^{[B]}_{\\tau})
    ```

    The reconciled sample paths allow for reconciled distributional forecasts:
    ```math
    (\mathbf{S}\mathbf{P}\hat{\mathbf{y}}^{[1]}_{\\tau}, \dots ,\mathbf{S}\mathbf{P}\hat{\mathbf{y}}^{[B]}_{\\tau})
    ```

    Args:
        S: np.array, summing matrix of size (`base`, `bottom`).
        P: np.array, reconciliation matrix of size (`bottom`, `base`).
        y_hat: Point forecasts values of size (`base`, `horizon`).
        y_insample: Insample values of size (`base`, `insample_size`).
        y_hat_insample: Insample point forecasts of size (`base`, `insample_size`).
        num_samples: int, number of bootstraped samples generated.
        seed: int, random seed for numpy generator's replicability.

    References:
        - [Puwasala Gamakumara Ph. D. dissertation. Monash University, Econometrics and Business Statistics (2020). "Probabilistic Forecast Reconciliation"](https://bridges.monash.edu/articles/thesis/Probabilistic_Forecast_Reconciliation_Theory_and_Applications/11869533)
        - [Panagiotelis A., Gamakumara P. Athanasopoulos G., and Hyndman R. J. (2022). "Probabilistic forecast reconciliation: Properties, evaluation and score optimisation". European Journal of Operational Research.](https://www.sciencedirect.com/science/article/pii/S0377221722006087)
    """

    def __init__(
        self,
        S: np.ndarray | sp.spmatrix,
        P: np.ndarray | sp.spmatrix,
        y_hat: np.ndarray,
        y_insample: np.ndarray,
        y_hat_insample: np.ndarray,
        num_samples: int = 100,
        seed: int = 0,
        W: np.ndarray | sp.spmatrix = None,
    ):
        self.S = S
        self.P = P
        self.W = W
        self.y_hat = y_hat
        self.y_insample = y_insample
        self.y_hat_insample = y_hat_insample
        self.num_samples = num_samples
        self.seed = seed

    def get_samples(self, num_samples: int):
        """Bootstrap Sample Reconciliation Method.

        Applies Bootstrap sample reconciliation method as defined by Gamakumara 2020.
        Generating independent sample paths and reconciling them with Bootstrap.

        Args:
            num_samples: int, number of samples generated from coherent distribution.

        Returns:
            samples: Coherent samples of size (`base`, `horizon`, `num_samples`).
        """
        residuals = self.y_insample - self.y_hat_insample
        h = self.y_hat.shape[1]

        # removing nas from residuals
        residuals = residuals[:, np.isnan(residuals).sum(axis=0) == 0]
        sample_idx = np.arange(residuals.shape[1] - h)
        rng = np.random.default_rng(self.seed)
        samples_idx = rng.choice(sample_idx, size=num_samples)
        samples = [self.y_hat + residuals[:, idx : (idx + h)] for idx in samples_idx]
        SP = self.S @ self.P
        samples = np.apply_along_axis(lambda path: SP @ path, axis=1, arr=samples)
        samples_np = np.stack(samples)

        # [samples, N, H] -> [N, H, samples]
        samples_np = samples_np.transpose((1, 2, 0))
        return samples_np

    def get_prediction_levels(self, res, level):
        """Adds reconciled forecast levels to results dictionary"""
        samples = self.get_samples(num_samples=self.num_samples)
        for lv in level:
            min_q = (100 - lv) / 200
            max_q = min_q + lv / 100
            res[f"lo-{lv}"] = np.quantile(samples, min_q, axis=2)
            res[f"hi-{lv}"] = np.quantile(samples, max_q, axis=2)
        return res

    def get_prediction_quantiles(self, res, quantiles):
        """Adds reconciled forecast quantiles to results dictionary"""
        samples = self.get_samples(num_samples=self.num_samples)

        # [Q, N, H] -> [N, H, Q]
        sample_quantiles = np.quantile(samples, quantiles, axis=2)
        res["quantiles"] = sample_quantiles.transpose((1, 2, 0))
        return res


class PERMBU:
    r"""PERMBU Probabilistic Reconciliation Class.

    The PERMBU method leverages empirical bottom-level marginal distributions
    with empirical copula functions (describing bottom-level dependencies) to
    generate the distribution of aggregate-level distributions using BottomUp
    reconciliation. The sample reordering technique in the PERMBU method reinjects
    multivariate dependencies into independent bottom-level samples. $\hat{\epsilon}_{i,t}$

    Algorithm:
    1.   For all series compute conditional marginals distributions.
    2.   Compute residuals $\hat{\epsilon}_{i,t}$ and obtain rank permutations.
    3.   Obtain K-sample from the bottom-level series predictions.
    4.   Apply recursively through the hierarchical structure:
        4.1.   For a given aggregate series $i$ and its children series:
        4.2.   Obtain children's empirical joint using sample reordering copula.
        4.2.   From the children's joint obtain the aggregate series's samples.

    Args:
        S (np.array): summing matrix of size (`base`, `bottom`).
        tags (dict[str, np.ndarray]): Each key is a level and each value its `S` indices.
        y_insample (np.array): Insample values of size (`base`, `insample_size`).
        y_hat_insample (np.array): Insample point forecasts of size (`base`, `insample_size`).
        sigmah (np.array): forecast standard dev. of size (`base`, `horizon`).
        num_samples (int): number of normal prediction samples generated. Default is None
        seed (int): random seed for numpy generator's replicability. Default is 0.

    References:
        - [Taieb, Souhaib Ben and Taylor, James W and Hyndman, Rob J. (2017). "Coherent probabilistic forecasts for hierarchical time series. International conference on machine learning ICML."](https://proceedings.mlr.press/v70/taieb17a.html)
    """

    def __init__(
        self,
        S: np.ndarray | sp.spmatrix,
        tags: dict[str, np.ndarray],
        y_hat: np.ndarray,
        y_insample: np.ndarray,
        y_hat_insample: np.ndarray,
        sigmah: np.ndarray,
        num_samples: int | None = None,
        seed: int = 0,
        P: np.ndarray | sp.spmatrix = None,
    ):
        # PERMBU only works for strictly hierarchical structures
        if not is_strictly_hierarchical(S, tags):
            raise ValueError(
                "PERMBU probabilistic reconciliation requires strictly hierarchical structures."
            )
        self.S = S
        self.P = P
        self.y_hat = y_hat
        self.y_insample = y_insample
        self.y_hat_insample = y_hat_insample
        self.sigmah = sigmah
        self.num_samples = num_samples
        self.seed = seed

    def _obtain_ranks(self, array):
        """Vector ranks

        Efficiently obtain vector ranks.
        Example `array=[4,2,7,1]` -> `ranks=[2, 1, 3, 0]`.

        Args:
            array (np.ndarray): Matrix with floats or integers on which the ranks will be computed on the second dimension.

        Returns:
            np.ndarray: Matrix with ranks along the second dimension.
        """
        temp = array.argsort(axis=1)
        ranks = np.empty_like(temp)
        a_range = np.arange(temp.shape[1])
        for i_row in range(temp.shape[0]):
            ranks[i_row, temp[i_row, :]] = a_range
        return ranks

    def _permutate_samples(self, samples, permutations):
        """Permutate Samples

        Applies efficient vectorized permutation on the samples.

        Args:
            samples (np.ndarray): Independent base samples.
            permutations (np.ndarray): Permutation ranks with which `samples` dependence will be restored see `_obtain_ranks`.

        Returns:
            np.ndarray: Permutated samples.
        """
        # Generate auxiliary and flat permutation indexes
        n_rows, n_cols = permutations.shape
        aux_row_idx = np.arange(n_rows)[:, None] * n_cols
        aux_row_idx = np.repeat(aux_row_idx, repeats=n_cols, axis=1)
        permutate_idxs = permutations.flatten() + aux_row_idx.flatten()

        # Apply flat permutation indexes and recover original shape
        permutated_samples = samples.flatten()
        permutated_samples = permutated_samples[permutate_idxs]
        permutated_samples = permutated_samples.reshape(n_rows, n_cols)
        return permutated_samples

    def _permutate_predictions(self, prediction_samples, permutations):
        """Permutate Prediction Samples

        Applies permutations to prediction_samples across the horizon.

        Args:
            prediction_samples (np.ndarray): Independent base prediction samples.
            permutations (np.ndarray): Permutation ranks with which `samples` dependence will be restored see `_obtain_ranks`.

        Returns:
            np.ndarray: Permutated prediction samples.
        """
        # Apply permutation throughout forecast horizon
        permutated_prediction_samples = prediction_samples.copy()

        _, n_horizon, _ = prediction_samples.shape
        for t in range(n_horizon):
            permutated_prediction_samples[:, t, :] = self._permutate_samples(
                prediction_samples[:, t, :], permutations
            )
        return permutated_prediction_samples

    def _nonzero_indexes_by_row(self, M):
        nonzeros = M.nonzero()[1]
        nnz_per_row = int(M[0, :].sum())
        return nonzeros.reshape(-1, nnz_per_row)

    def get_samples(self, num_samples: int | None = None):
        """PERMBU Sample Reconciliation Method.

        Applies PERMBU reconciliation method as defined by Taieb et. al 2017.
        Generating independent base prediction samples, restoring its multivariate
        dependence using estimated copula with reordering and applying the BottomUp
        aggregation to the new samples.

        Args:
            num_samples (int): number of samples generated from coherent distribution.

        Returns:
            samples (np.ndarray): Coherent samples of size (`base`, `horizon`, `num_samples`).
        """
        # Compute residuals and rank permutations
        residuals = self.y_insample - self.y_hat_insample
        residuals = residuals[:, np.isnan(residuals).sum(axis=0) == 0]

        # Sample h step-ahead base marginal distributions
        if num_samples is None:
            num_samples = residuals.shape[1]

        # Expand residuals to match num_samples [(a,b),T] -> [(a,b),num_samples]
        rng = np.random.default_rng(self.seed)
        if num_samples > residuals.shape[1]:
            residuals_idxs = rng.choice(residuals.shape[1], size=num_samples)
        else:
            residuals_idxs = rng.choice(
                residuals.shape[1], size=num_samples, replace=False
            )
        residuals = residuals[:, residuals_idxs]
        rank_permutations = self._obtain_ranks(residuals)

        n_series, n_horizon = self.y_hat.shape

        base_samples = np.array(
            [
                rng.normal(loc=m, scale=s, size=num_samples)
                for m, s in zip(self.y_hat.flatten(), self.sigmah.flatten(), strict=False)
            ]
        )
        base_samples = base_samples.reshape(n_series, n_horizon, num_samples)

        # Initialize PERMBU utility
        rec_samples = base_samples.copy()
        try:
            encoder = OneHotEncoder(sparse_output=False, dtype=np.float64)
        except TypeError:
            encoder = OneHotEncoder(sparse=False, dtype=np.float64)
        hier_links = self._nonzero_indexes_by_row(self.S.T)

        # BottomUp hierarchy traversing
        hier_levels = hier_links.shape[1] - 1
        for level_idx in reversed(range(hier_levels)):
            # Obtain aggregation matrix from parent/children links
            children_links = np.unique(hier_links[:, level_idx : level_idx + 2], axis=0)
            children_idxs = np.unique(children_links[:, 1])
            parent_idxs = np.unique(children_links[:, 0])
            Agg = encoder.fit_transform(children_links).T
            Agg = Agg[: len(parent_idxs), :]

            # Permute children_samples for each prediction step
            children_permutations = rank_permutations[children_idxs, :]
            children_samples = rec_samples[children_idxs, :, :]
            children_samples = self._permutate_predictions(
                prediction_samples=children_samples, permutations=children_permutations
            )

            # Overwrite hier_samples with BottomUp aggregation
            # and randomly shuffle parent predictions after aggregation
            parent_samples = np.einsum("ab,bhs->ahs", Agg, children_samples)
            random_permutation = np.array(
                [
                    rng.permutation(np.arange(num_samples))
                    for serie in range(len(parent_samples))
                ]
            )
            parent_samples = self._permutate_predictions(
                prediction_samples=parent_samples, permutations=random_permutation
            )

            rec_samples[parent_idxs, :, :] = parent_samples
        return rec_samples

    def get_prediction_levels(self, res, level):
        """Adds reconciled forecast levels to results dictionary"""
        samples = self.get_samples(num_samples=self.num_samples)
        for lv in level:
            min_q = (100 - lv) / 200
            max_q = min_q + lv / 100
            res[f"lo-{lv}"] = np.quantile(samples, min_q, axis=2)
            res[f"hi-{lv}"] = np.quantile(samples, max_q, axis=2)
        return res

    def get_prediction_quantiles(self, res, quantiles):
        """Adds reconciled forecast quantiles to results dictionary"""
        samples = self.get_samples(num_samples=self.num_samples)

        # [Q, N, H] -> [N, H, Q]
        sample_quantiles = np.quantile(samples, quantiles, axis=2)
        res["quantiles"] = sample_quantiles.transpose((1, 2, 0))
        return res


class Conformal:
    r"""Conformal Prediction for Hierarchical Reconciliation Class.

    This class implements distribution-free prediction intervals with guaranteed
    finite-sample coverage for hierarchically reconciled forecasts. Unlike the
    Normality method which assumes Gaussian distributions, conformal prediction
    provides valid coverage without distributional assumptions.

    The method follows the Split Conformal Prediction (SCP) framework with
    hierarchical reconciliation, as proposed by Principato et al. (2024).

    **Algorithm (Component-wise SCP with reconciliation):**
    1. Split data into training (fit forecaster) and calibration sets
    2. Compute reconciled predictions: $\tilde{y} = SP \cdot \hat{y}$
    3. Compute signed non-conformity scores: $s_t = y_t - \tilde{y}_t$ on calibration data
    4. For each component i, order scores and compute quantiles
    5. Prediction intervals use empirical quantiles of non-conformity scores

    **Coverage Guarantee:**
    For a calibration set of size n and confidence level L (e.g., 90), the theoretical
    coverage is at least $(L/100) \cdot n/(n+1)$.

    **Important:**
    The coverage guarantee assumes a proper calibration set that is independent from
    the training data. When using in-sample residuals (i.e., `y_cal` and `y_hat_cal`
    from the same data used to train the forecasting model), this assumption is violated.
    In-sample residuals typically underestimate true forecast errors, which may result
    in prediction intervals that are narrower than their nominal coverage suggests.
    For valid coverage guarantees, use a held-out calibration set.

    Args:
        S (Union[np.ndarray, sp.spmatrix]): Summing matrix of size (`n_series`, `n_bottom`).
        P (Union[np.ndarray, sp.spmatrix]): Reconciliation matrix of size (`n_bottom`, `n_series`).
        y_hat (np.ndarray): Point forecasts of size (`n_series`, `horizon`).
        y_cal (np.ndarray): Calibration actual values of size (`n_series`, `n_cal`).
        y_hat_cal (np.ndarray): Calibration predictions of size (`n_series`, `n_cal`).
        seed (int, optional): Random seed for numpy generator's replicability. Default is 0.

    Raises:
        ValueError: If calibration set is empty or has fewer than 2 observations.
        ValueError: If y_cal and y_hat_cal shapes don't match.

    References:
        - [Principato G., Stoltz G., Amara-Ouali Y., Goude Y., Hamrouche B., Poggi J-M. (2024).
          "Conformal Prediction for Hierarchical Data". arXiv:2411.13479](https://arxiv.org/abs/2411.13479)
        - [Lei J., G'Sell M., Rinaldo A., Tibshirani R., Wasserman L. (2018).
          "Distribution-Free Predictive Inference For Regression".
          Journal of the American Statistical Association.](https://doi.org/10.1080/01621459.2017.1307116)

    Examples:
        >>> conformal = Conformal(
        ...     S=S, P=P, y_hat=y_hat,
        ...     y_cal=y_insample, y_hat_cal=y_hat_insample,
        ... )
        >>> samples = conformal.get_samples(num_samples=100)
        >>> res = {"mean": y_reconciled}
        >>> res = conformal.get_prediction_levels(res, level=[90, 95])
    """

    def __init__(
        self,
        S: np.ndarray | sp.spmatrix,
        P: np.ndarray | sp.spmatrix,
        y_hat: np.ndarray,
        y_cal: np.ndarray,
        y_hat_cal: np.ndarray,
        seed: int = 0,
    ):
        if y_cal.ndim != 2:
            raise ValueError(
                f"y_cal must be a 2D array of shape (n_series, n_cal), "
                f"got {y_cal.ndim}D array with shape {y_cal.shape}."
            )

        if y_hat_cal.ndim != 2:
            raise ValueError(
                f"y_hat_cal must be a 2D array of shape (n_series, n_cal), "
                f"got {y_hat_cal.ndim}D array with shape {y_hat_cal.shape}."
            )

        if y_cal.shape != y_hat_cal.shape:
            raise ValueError(
                f"y_cal and y_hat_cal shapes must match. "
                f"Got y_cal: {y_cal.shape}, y_hat_cal: {y_hat_cal.shape}."
            )

        n_series = S.shape[0]
        if y_cal.shape[0] != n_series:
            raise ValueError(
                f"y_cal first dimension ({y_cal.shape[0]}) must match "
                f"number of series in S ({n_series})."
            )

        n_cal = y_cal.shape[1]
        if n_cal < 2:
            raise ValueError(
                f"Calibration set has only {n_cal} observation(s). "
                "At least 2 observations are required for conformal prediction."
            )

        nan_mask = np.isnan(y_cal) | np.isnan(y_hat_cal)
        valid_cols = ~np.any(nan_mask, axis=0)
        n_valid = np.sum(valid_cols)

        if n_valid < 2:
            raise ValueError(
                f"Only {n_valid} valid (non-NaN) calibration observations. "
                "At least 2 valid observations are required."
            )

        if n_valid < n_cal:
            warnings.warn(
                f"Calibration set reduced from {n_cal} to {n_valid} observations "
                "due to NaN values.",
                UserWarning,
                stacklevel=2,
            )

        self.S = S
        self.P = P
        self.y_hat = y_hat
        self.seed = seed

        if isinstance(P, sp.linalg.LinearOperator) and sp.issparse(S):
            self.SP = sp.linalg.aslinearoperator(self.S) @ self.P
        else:
            self.SP = self.S @ self.P

        y_cal_clean = y_cal[:, valid_cols]
        y_hat_cal_clean = y_hat_cal[:, valid_cols]
        self.n_cal = n_valid

        y_hat_cal_rec = self.SP @ y_hat_cal_clean
        self.scores = y_cal_clean - y_hat_cal_rec
        self.sorted_scores = np.sort(self.scores, axis=1)
        self.y_hat_rec = self.SP @ self.y_hat

    def get_samples(self, num_samples: int) -> np.ndarray:
        """Generate samples from the conformal prediction distribution.

        Uses bootstrap resampling of non-conformity scores to generate
        coherent samples.

        Args:
            num_samples (int): Number of samples to generate. Must be positive.

        Returns:
            np.ndarray: Samples of size (n_series, horizon, num_samples).

        Raises:
            ValueError: If num_samples is not positive.
        """
        if num_samples <= 0:
            raise ValueError(
                f"num_samples must be a positive integer, got {num_samples}."
            )

        rng = np.random.default_rng(self.seed)
        n_series, n_horizon = self.y_hat_rec.shape

        samples = np.empty((n_series, n_horizon, num_samples))

        for t in range(n_horizon):
            sample_idx = rng.choice(self.n_cal, size=num_samples)
            sampled_scores = self.scores[:, sample_idx]
            samples[:, t, :] = self.y_hat_rec[:, t, np.newaxis] + sampled_scores

        return samples

    def get_prediction_levels(self, res: dict, level: list) -> dict:
        """Add reconciled forecast levels to results dictionary.

        Args:
            res (dict): Results dictionary to update.
            level (list): Confidence levels, e.g., [90, 95] for 90% and 95%.

        Returns:
            dict: Updated results dictionary with 'lo-{lv}' and 'hi-{lv}' keys.
        """
        for lv in level:
            alpha_lv = 1 - lv / 100
            n = self.n_cal

            k_lo = int(np.floor((n + 1) * (alpha_lv / 2)))
            k_hi = int(np.ceil((n + 1) * (1 - alpha_lv / 2)))
            k_lo = max(0, min(k_lo, n - 1))
            k_hi = max(0, min(k_hi - 1, n - 1))

            lo_bound = self.sorted_scores[:, k_lo]
            hi_bound = self.sorted_scores[:, k_hi]

            res[f"lo-{lv}"] = self.y_hat_rec + lo_bound[:, np.newaxis]
            res[f"hi-{lv}"] = self.y_hat_rec + hi_bound[:, np.newaxis]

        return res

    def get_prediction_quantiles(self, res: dict, quantiles: np.ndarray) -> dict:
        """Add reconciled forecast quantiles to results dictionary.

        Args:
            res (dict): Results dictionary to update.
            quantiles (np.ndarray): Quantiles to compute, e.g., [0.025, 0.5, 0.975].

        Returns:
            dict: Updated results dictionary with 'quantiles' key.
        """
        quantiles = np.asarray(quantiles)
        n_series, n_horizon = self.y_hat_rec.shape
        n_quantiles = len(quantiles)

        score_quantiles = np.quantile(self.scores, quantiles, axis=1)

        result_quantiles = np.empty((n_series, n_horizon, n_quantiles))
        for q_idx in range(n_quantiles):
            result_quantiles[:, :, q_idx] = (
                self.y_hat_rec + score_quantiles[q_idx, :, np.newaxis]
            )

        res["quantiles"] = result_quantiles
        return res




__all__ = ['evaluate']


from collections.abc import Callable

import narwhals.stable.v2 as nw
import numpy as np
import utilsforecast.evaluation as ufe
from narwhals.typing import FrameT


def evaluate(
    df: FrameT,
    metrics: list[Callable],
    tags: dict[str, np.ndarray],
    models: list[str] | None = None,
    train_df: FrameT | None = None,
    level: list[int] | None = None,
    id_col: str = "unique_id",
    time_col: str = "ds",
    target_col: str = "y",
    agg_fn: str | None = "mean",
    benchmark: str | None = None,
) -> FrameT:
    """Evaluate hierarchical forecast using different metrics.

    Args:
        df (pandas, polars, dask or spark DataFrame): Forecasts to evaluate. Must have `id_col`, `time_col`, `target_col` and models' predictions.
        metrics (list of callable): Functions with arguments `df`, `models`, `id_col`, `target_col` and optionally `train_df`.
        tags (dict): Each key is a level in the hierarchy and its value contains tags associated to that level. Each key is a level in the hierarchy and its value contains tags associated to that level.
        models (list of str, optional): Names of the models to evaluate. If `None` will use every column in the dataframe after removing id, time and target.
        train_df (pandas, polars, dask or spark DataFrame, optional): Training set. Used to evaluate metrics such as `mase`.
        level (list of int, optional): Prediction interval levels. Used to compute losses that rely on quantiles.
        id_col (str): Column that identifies each serie.
        time_col (str): Column that identifies each timestep, its values can be timestamps or integers.
        target_col (str): Column that contains the target.
        agg_fn (str, optional): Statistic to compute on the scores by id to reduce them to a single number.
        benchmark (str, optional): If passed, evaluators are scaled by the error of this benchmark model.

    Returns:
        pandas, polars DataFrame: Metrics with one row per (id, metric) combination and one column per model. If `agg_fn` is not `None`, there is only one row per metric.
    """
    # Check benchmark in columns
    if benchmark is not None:
        if benchmark not in df.columns:
            raise ValueError(f"Benchmark model '{benchmark}' not found in df")
        model_cols = None

    df_nw = nw.from_native(df)
    if train_df is not None:
        train_nw = nw.from_native(train_df)
    tag_scores = []
    tags_ = {**tags, "Overall": np.concatenate(list(tags.values()))}

    eps = np.finfo(np.float32).eps

    for tag, tag_ids in tags_.items():
        df_tag = df_nw.filter(nw.col(id_col).is_in(tag_ids)).to_native()
        if train_df is not None:
            train_nw_tag = train_nw.filter(nw.col(id_col).is_in(tag_ids)).to_native()
        else:
            train_nw_tag = None

        df_score = ufe.evaluate(
            df=df_tag,
            metrics=metrics,
            models=models,
            train_df=train_nw_tag,
            level=level,
            id_col=id_col,
            time_col=time_col,
            target_col=target_col,
            agg_fn=agg_fn,
        )
        df_score_nw = nw.from_native(df_score)
        if benchmark is not None:
            if model_cols is None:
                model_cols = [
                    c for c in df_score_nw.columns if c not in [id_col, "metric"]
                ]
            # NB: the below assumes benchmark loss is always positive, which is a reasonable assumption
            df_score_nw = df_score_nw.with_columns(
                nw.col(model_cols) / nw.col(benchmark).clip(eps)
            )
            df_score_nw = df_score_nw.with_columns(nw.col("metric") + "-scaled")

        df_score_nw = df_score_nw.select(nw.lit(tag).alias("level"), nw.all())
        tag_scores.append(df_score_nw)

    tag_scores_nw = nw.concat(tag_scores)
    tag_scores_nw = nw.maybe_reset_index(tag_scores_nw)

    df_eval = tag_scores_nw.to_native()

    return df_eval



import warnings

import narwhals.stable.v2 as nw
import numpy as np
import pytest
import scipy.sparse as sp

from hierarchicalforecast.core import HierarchicalReconciliation
from hierarchicalforecast.methods import BottomUp
from hierarchicalforecast.probabilistic_methods import (
    PERMBU,
    Bootstrap,
    Conformal,
    CovarianceType,
    Normality,
)


@pytest.fixture
def test_data():
    """Fixture to provide test data for hierarchical forecasting tests."""
    S = np.array(
        [
            [1.0, 1.0, 1.0, 1.0],
            [1.0, 1.0, 0.0, 0.0],
            [0.0, 0.0, 1.0, 1.0],
            [1.0, 0.0, 0.0, 0.0],
            [0.0, 1.0, 0.0, 0.0],
            [0.0, 0.0, 1.0, 0.0],
            [0.0, 0.0, 0.0, 1.0],
        ]
    )
    h = 2
    _y = np.array([10.0, 5.0, 4.0, 2.0, 1.0])
    y_bottom = np.vstack([i * _y for i in range(1, 5)])
    y_hat_bottom_insample = np.roll(y_bottom, 1)
    y_hat_bottom_insample[:, 0] = np.nan
    y_hat_bottom = np.vstack([i * np.ones(h) for i in range(1, 5)])
    tags = {"level1": np.array([0]), "level2": np.array([1, 2]), "level3": np.array([3, 4, 5, 6])}

    y_base = S @ y_bottom
    y_hat_base = S @ y_hat_bottom
    y_hat_base_insample = S @ y_hat_bottom_insample
    sigma = np.nansum((y_base - y_hat_base_insample) ** 2, axis=1) / (
        y_base.shape[1] - 1
    )
    sigma = np.sqrt(sigma)
    sigmah = sigma[:, None] * np.sqrt(
        np.vstack([np.arange(1, h + 1) for _ in range(y_base.shape[0])])
    )
    noise = np.random.normal(scale=sigmah)
    y_test = y_hat_base + noise

    return {
        "S": S,
        "h": h,
        "y_bottom": y_bottom,
        "y_hat_bottom": y_hat_bottom,
        "y_hat_bottom_insample": y_hat_bottom_insample,
        "tags": tags,
        "y_base": y_base,
        "y_hat_base": y_hat_base,
        "y_hat_base_insample": y_hat_base_insample,
        "sigmah": sigmah,
        "y_test": y_test,
    }


@pytest.fixture
def samplers(test_data):
    """Fixture to provide samplers for testing."""
    cls_bottom_up = BottomUp()
    P, W = cls_bottom_up._get_PW_matrices(
        S=test_data["S"]    )

    normality_sampler = Normality(
        S=test_data["S"],
        P=P,
        W=W,
        y_hat=test_data["y_hat_base"],
        sigmah=test_data["sigmah"],
    )
    bootstrap_sampler = Bootstrap(
        S=test_data["S"],
        P=P,
        W=W,
        y_hat=test_data["y_hat_base"],
        y_insample=test_data["y_base"],
        y_hat_insample=test_data["y_hat_base_insample"],
        num_samples=1_000,
    )
    empty_bootstrap_sampler = Bootstrap(
        S=test_data["S"],
        P=P,
        W=W,
        y_hat=test_data["y_hat_base"],
        y_insample=test_data["y_base"],
        y_hat_insample=test_data["y_base"],
        num_samples=1_000,
    )
    permbu_sampler = PERMBU(
        S=test_data["S"],
        P=P,
        tags=test_data["tags"],
        y_hat=test_data["y_hat_base"],
        y_insample=test_data["y_base"],
        y_hat_insample=test_data["y_hat_base_insample"],
        sigmah=test_data["sigmah"],
    )
    empty_permbu_sampler = PERMBU(
        S=test_data["S"],
        P=P,
        tags=test_data["tags"],
        y_hat=test_data["y_hat_base"],
        y_insample=test_data["y_base"],
        y_hat_insample=test_data["y_base"],
        sigmah=test_data["sigmah"],
    )

    return {
        "normality_sampler": normality_sampler,
        "bootstrap_sampler": bootstrap_sampler,
        "empty_bootstrap_sampler": empty_bootstrap_sampler,
        "permbu_sampler": permbu_sampler,
        "empty_permbu_sampler": empty_permbu_sampler,
    }


# Tests for Normality covariance_type parameter (Issue #99)
class TestNormalityCovarianceType:
    """Tests for Normality covariance_type parameter."""

    # Basic functionality tests

    def test_normality_diagonal_covariance(self, test_data):
        """Test Normality with diagonal covariance (default, backward compat)."""
        cls_bottom_up = BottomUp()
        P, W = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )

        # Default (diagonal) should work without residuals
        normality_diag = Normality(
            S=test_data["S"],
            P=P,
            W=W,
            y_hat=test_data["y_hat_base"],
            sigmah=test_data["sigmah"],
            covariance_type="diagonal",
        )
        samples = normality_diag.get_samples(num_samples=50)
        assert samples.shape == (test_data["S"].shape[0], test_data["h"], 50)

    def test_normality_full_covariance(self, test_data):
        """Test Normality with full empirical covariance."""
        cls_bottom_up = BottomUp()
        P, W = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )

        # Compute residuals from insample data
        residuals = test_data["y_base"] - test_data["y_hat_base_insample"]

        normality_full = Normality(
            S=test_data["S"],
            P=P,
            y_hat=test_data["y_hat_base"],
            sigmah=test_data["sigmah"],
            covariance_type="full",
            residuals=residuals,
        )
        samples = normality_full.get_samples(num_samples=50)
        assert samples.shape == (test_data["S"].shape[0], test_data["h"], 50)

    def test_normality_shrink_covariance(self, test_data):
        """Test Normality with shrinkage covariance (Schäfer-Strimmer)."""
        cls_bottom_up = BottomUp()
        P, W = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )

        # Compute residuals from insample data
        residuals = test_data["y_base"] - test_data["y_hat_base_insample"]

        normality_shrink = Normality(
            S=test_data["S"],
            P=P,
            y_hat=test_data["y_hat_base"],
            sigmah=test_data["sigmah"],
            covariance_type="shrink",
            residuals=residuals,
            shrinkage_ridge=2e-8,
        )
        samples = normality_shrink.get_samples(num_samples=50)
        assert samples.shape == (test_data["S"].shape[0], test_data["h"], 50)

    def test_normality_w_optional_for_full(self, test_data):
        """Test that W is optional when covariance_type is 'full'."""
        cls_bottom_up = BottomUp()
        P, _ = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )
        residuals = test_data["y_base"] - test_data["y_hat_base_insample"]

        # Should work without W
        normality = Normality(
            S=test_data["S"],
            P=P,
            y_hat=test_data["y_hat_base"],
            sigmah=test_data["sigmah"],
            covariance_type="full",
            residuals=residuals,
        )
        samples = normality.get_samples(num_samples=50)
        assert samples.shape == (test_data["S"].shape[0], test_data["h"], 50)

    def test_normality_w_optional_for_shrink(self, test_data):
        """Test that W is optional when covariance_type is 'shrink'."""
        cls_bottom_up = BottomUp()
        P, _ = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )
        residuals = test_data["y_base"] - test_data["y_hat_base_insample"]

        # Should work without W
        normality = Normality(
            S=test_data["S"],
            P=P,
            y_hat=test_data["y_hat_base"],
            sigmah=test_data["sigmah"],
            covariance_type="shrink",
            residuals=residuals,
        )
        samples = normality.get_samples(num_samples=50)
        assert samples.shape == (test_data["S"].shape[0], test_data["h"], 50)

    # Enum and case sensitivity tests

    def test_normality_covariance_type_enum(self, test_data):
        """Test that CovarianceType enum works."""
        cls_bottom_up = BottomUp()
        P, W = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )
        residuals = test_data["y_base"] - test_data["y_hat_base_insample"]

        # Test with enum
        normality = Normality(
            S=test_data["S"],
            P=P,
            y_hat=test_data["y_hat_base"],
            sigmah=test_data["sigmah"],
            covariance_type=CovarianceType.SHRINK,
            residuals=residuals,
        )
        samples = normality.get_samples(num_samples=50)
        assert samples.shape == (test_data["S"].shape[0], test_data["h"], 50)
        assert normality.covariance_type == CovarianceType.SHRINK

    def test_normality_case_insensitive_covariance_type(self, test_data):
        """Test that covariance_type is case insensitive."""
        cls_bottom_up = BottomUp()
        P, W = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )
        residuals = test_data["y_base"] - test_data["y_hat_base_insample"]

        # Test various cases
        for cov_type in ["FULL", "Full", "fUlL"]:
            normality = Normality(
                S=test_data["S"],
                P=P,
                y_hat=test_data["y_hat_base"],
                sigmah=test_data["sigmah"],
                covariance_type=cov_type,
                residuals=residuals,
            )
            assert normality.covariance_type == CovarianceType.FULL

    # Validation error tests

    def test_normality_invalid_covariance_type(self, test_data):
        """Test that invalid covariance_type raises ValueError."""
        cls_bottom_up = BottomUp()
        P, W = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )

        with pytest.raises(ValueError) as exc_info:
            Normality(
                S=test_data["S"],
                P=P,
                W=W,
                y_hat=test_data["y_hat_base"],
                sigmah=test_data["sigmah"],
                covariance_type="invalid_type",
            )
        assert "Unknown covariance_type" in str(exc_info.value)

    def test_normality_invalid_type_for_covariance_type(self, test_data):
        """Test that non-string/non-enum covariance_type raises ValueError."""
        cls_bottom_up = BottomUp()
        P, W = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )

        with pytest.raises(ValueError) as exc_info:
            Normality(
                S=test_data["S"],
                P=P,
                W=W,
                y_hat=test_data["y_hat_base"],
                sigmah=test_data["sigmah"],
                covariance_type=123,  # Invalid type
            )
        assert "must be a string or CovarianceType enum" in str(exc_info.value)

    def test_normality_full_requires_residuals(self, test_data):
        """Test that full covariance requires residuals."""
        cls_bottom_up = BottomUp()
        P, W = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )

        with pytest.raises(ValueError) as exc_info:
            Normality(
                S=test_data["S"],
                P=P,
                W=W,
                y_hat=test_data["y_hat_base"],
                sigmah=test_data["sigmah"],
                covariance_type="full",
            )
        assert "requires `residuals` parameter" in str(exc_info.value)

    def test_normality_shrink_requires_residuals(self, test_data):
        """Test that shrink covariance requires residuals."""
        cls_bottom_up = BottomUp()
        P, W = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )

        with pytest.raises(ValueError) as exc_info:
            Normality(
                S=test_data["S"],
                P=P,
                W=W,
                y_hat=test_data["y_hat_base"],
                sigmah=test_data["sigmah"],
                covariance_type="shrink",
            )
        assert "requires `residuals` parameter" in str(exc_info.value)

    def test_normality_diagonal_requires_w(self, test_data):
        """Test that diagonal covariance requires W."""
        cls_bottom_up = BottomUp()
        P, _ = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )

        with pytest.raises(ValueError) as exc_info:
            Normality(
                S=test_data["S"],
                P=P,
                y_hat=test_data["y_hat_base"],
                sigmah=test_data["sigmah"],
                covariance_type="diagonal",
            )
        assert "covariance_type='diagonal' requires `W` parameter" in str(
            exc_info.value
        )

    # Residuals shape validation tests

    def test_normality_residuals_wrong_shape_1d(self, test_data):
        """Test that 1D residuals raises ValueError."""
        cls_bottom_up = BottomUp()
        P, _ = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )
        residuals_1d = np.array([1.0, 2.0, 3.0])

        with pytest.raises(ValueError) as exc_info:
            Normality(
                S=test_data["S"],
                P=P,
                y_hat=test_data["y_hat_base"],
                sigmah=test_data["sigmah"],
                covariance_type="full",
                residuals=residuals_1d,
            )
        assert "must be a 2D array" in str(exc_info.value)

    def test_normality_residuals_wrong_n_series(self, test_data):
        """Test that residuals with wrong number of series raises ValueError."""
        cls_bottom_up = BottomUp()
        P, _ = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )

        residuals_wrong = np.random.randn(5, 10)

        with pytest.raises(ValueError) as exc_info:
            Normality(
                S=test_data["S"],
                P=P,
                y_hat=test_data["y_hat_base"],
                sigmah=test_data["sigmah"],
                covariance_type="full",
                residuals=residuals_wrong,
            )
        assert "residuals shape mismatch" in str(exc_info.value)

    def test_normality_residuals_empty(self, test_data):
        """Test that empty residuals raises ValueError."""
        cls_bottom_up = BottomUp()
        P, _ = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )
        residuals_empty = np.empty((7, 0))

        with pytest.raises(ValueError) as exc_info:
            Normality(
                S=test_data["S"],
                P=P,
                y_hat=test_data["y_hat_base"],
                sigmah=test_data["sigmah"],
                covariance_type="full",
                residuals=residuals_empty,
            )
        assert "residuals is empty" in str(exc_info.value)

    def test_normality_residuals_single_observation(self, test_data):
        """Test that residuals with single observation raises ValueError."""
        cls_bottom_up = BottomUp()
        P, _ = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )
        residuals_single = np.random.randn(7, 1)

        with pytest.raises(ValueError) as exc_info:
            Normality(
                S=test_data["S"],
                P=P,
                y_hat=test_data["y_hat_base"],
                sigmah=test_data["sigmah"],
                covariance_type="full",
                residuals=residuals_single,
            )
        assert "At least 2 observations are required" in str(exc_info.value)

    # NaN handling tests

    def test_normality_residuals_all_nan_series(self, test_data):
        """Test that all-NaN series in residuals raises ValueError."""
        cls_bottom_up = BottomUp()
        P, _ = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )
        residuals = test_data["y_base"] - test_data["y_hat_base_insample"]
        # Make first series all NaN
        residuals[0, :] = np.nan

        with pytest.raises(ValueError) as exc_info:
            Normality(
                S=test_data["S"],
                P=P,
                y_hat=test_data["y_hat_base"],
                sigmah=test_data["sigmah"],
                covariance_type="full",
                residuals=residuals,
            )
        assert "all NaN values" in str(exc_info.value)

    def test_normality_residuals_insufficient_non_nan(self, test_data):
        """Test that series with <2 non-NaN observations raises ValueError."""
        cls_bottom_up = BottomUp()
        P, _ = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )
        residuals = test_data["y_base"] - test_data["y_hat_base_insample"]
        # Make first series have only 1 non-NaN value
        residuals[0, :] = np.nan
        residuals[0, 0] = 1.0

        with pytest.raises(ValueError) as exc_info:
            Normality(
                S=test_data["S"],
                P=P,
                y_hat=test_data["y_hat_base"],
                sigmah=test_data["sigmah"],
                covariance_type="full",
                residuals=residuals,
            )
        assert "fewer than 2 non-NaN observations" in str(exc_info.value)

    def test_normality_residuals_with_some_nans(self, test_data):
        """Test that residuals with some NaNs work correctly."""
        cls_bottom_up = BottomUp()
        P, _ = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )
        residuals = test_data["y_base"] - test_data["y_hat_base_insample"]
        # residuals already has NaN from y_hat_base_insample

        # Should work for both full and shrink
        for cov_type in ["full", "shrink"]:
            normality = Normality(
                S=test_data["S"],
                P=P,
                y_hat=test_data["y_hat_base"],
                sigmah=test_data["sigmah"],
                covariance_type=cov_type,
                residuals=residuals,
            )
            samples = normality.get_samples(num_samples=50)
            assert samples.shape == (test_data["S"].shape[0], test_data["h"], 50)

    # Warning tests

    def test_normality_warns_w_ignored(self, test_data):
        """Test that warning is issued when W is provided but ignored."""
        cls_bottom_up = BottomUp()
        P, W = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )
        residuals = test_data["y_base"] - test_data["y_hat_base_insample"]

        with pytest.warns(UserWarning, match="W parameter is ignored"):
            Normality(
                S=test_data["S"],
                P=P,
                W=W,  # Providing W when using 'full'
                y_hat=test_data["y_hat_base"],
                sigmah=test_data["sigmah"],
                covariance_type="full",
                residuals=residuals,
            )

    def test_normality_warns_shrinkage_ridge_ignored(self, test_data):
        """Test that warning is issued when shrinkage_ridge is provided but not used."""
        cls_bottom_up = BottomUp()
        P, W = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )

        with pytest.warns(UserWarning, match="shrinkage_ridge parameter is only used"):
            Normality(
                S=test_data["S"],
                P=P,
                W=W,
                y_hat=test_data["y_hat_base"],
                sigmah=test_data["sigmah"],
                covariance_type="diagonal",
                shrinkage_ridge=1e-6,  # Non-default value
            )

    def test_normality_warns_n_series_gt_n_obs(self, test_data):
        """Test warning when n_series > n_observations for full covariance."""
        cls_bottom_up = BottomUp()
        P, _ = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )
        # Create residuals with fewer observations than series
        n_series = test_data["S"].shape[0]  # 7 series
        residuals = np.random.randn(n_series, 3)  # Only 3 observations

        with pytest.warns(UserWarning, match="non-positive-definite"):
            Normality(
                S=test_data["S"],
                P=P,
                y_hat=test_data["y_hat_base"],
                sigmah=test_data["sigmah"],
                covariance_type="full",
                residuals=residuals,
            )

    # Zero/near-zero variance tests

    def test_normality_warns_zero_variance(self, test_data):
        """Test warning when series has zero variance."""
        cls_bottom_up = BottomUp()
        P, _ = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )
        n_series = test_data["S"].shape[0]
        residuals = np.random.randn(n_series, 10)
        # Make first series constant (zero variance)
        residuals[0, :] = 5.0

        with pytest.warns(UserWarning, match="zero or near-zero variance"):
            Normality(
                S=test_data["S"],
                P=P,
                y_hat=test_data["y_hat_base"],
                sigmah=test_data["sigmah"],
                covariance_type="full",
                residuals=residuals,
            )

    def test_normality_handles_zero_variance_gracefully(self, test_data):
        """Test that zero variance series doesn't crash and produces valid samples."""
        cls_bottom_up = BottomUp()
        P, _ = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )
        n_series = test_data["S"].shape[0]
        residuals = np.random.randn(n_series, 10)
        residuals[0, :] = 5.0  # Constant series

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            normality = Normality(
                S=test_data["S"],
                P=P,
                y_hat=test_data["y_hat_base"],
                sigmah=test_data["sigmah"],
                covariance_type="full",
                residuals=residuals,
            )
            samples = normality.get_samples(num_samples=50)

        assert samples.shape == (n_series, test_data["h"], 50)
        assert np.all(np.isfinite(samples))

    # W matrix validation tests

    def test_normality_w_nan_diagonal(self, test_data):
        """Test that W with NaN diagonal raises ValueError."""
        cls_bottom_up = BottomUp()
        P, W = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )
        W_with_nan = W.copy()
        W_with_nan[0, 0] = np.nan

        with pytest.raises(ValueError) as exc_info:
            Normality(
                S=test_data["S"],
                P=P,
                W=W_with_nan,
                y_hat=test_data["y_hat_base"],
                sigmah=test_data["sigmah"],
                covariance_type="diagonal",
            )
        assert "NaN values" in str(exc_info.value)

    def test_normality_w_non_positive_diagonal(self, test_data):
        """Test that W with non-positive diagonal raises ValueError."""
        cls_bottom_up = BottomUp()
        P, W = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )
        W_with_zero = W.copy()
        W_with_zero[0, 0] = 0.0

        with pytest.raises(ValueError) as exc_info:
            Normality(
                S=test_data["S"],
                P=P,
                W=W_with_zero,
                y_hat=test_data["y_hat_base"],
                sigmah=test_data["sigmah"],
                covariance_type="diagonal",
            )
        assert "non-positive diagonal" in str(exc_info.value)

    # n_series > n_observations tests (non-PSD risk)

    def test_normality_shrink_handles_high_dimensional_data(self, test_data):
        """Test that shrink covariance handles n_series > n_observations well."""
        cls_bottom_up = BottomUp()
        P, _ = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )
        n_series = test_data["S"].shape[0]
        # Create high-dimensional case: more series than observations
        residuals = np.random.randn(n_series, 3)

        # Shrink should work without warning about non-PSD
        normality = Normality(
            S=test_data["S"],
            P=P,
            y_hat=test_data["y_hat_base"],
            sigmah=test_data["sigmah"],
            covariance_type="shrink",
            residuals=residuals,
        )
        samples = normality.get_samples(num_samples=50)
        assert samples.shape == (n_series, test_data["h"], 50)
        assert np.all(np.isfinite(samples))

    # Statistical validation tests

    def test_normality_samples_have_correct_mean(self, test_data):
        """Test that samples have approximately correct mean."""
        cls_bottom_up = BottomUp()
        P, W = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )

        normality = Normality(
            S=test_data["S"],
            P=P,
            W=W,
            y_hat=test_data["y_hat_base"],
            sigmah=test_data["sigmah"],
            covariance_type="diagonal",
            seed=42,
        )

        # Generate many samples for statistical test
        num_samples = 50000
        samples = normality.get_samples(num_samples=num_samples)
        sample_means = samples.mean(axis=2)

        # Expected mean is SP @ y_hat
        expected_means = normality.SP @ test_data["y_hat_base"]

        # Check that sample means are close to expected means
        # Using relative tolerance due to different scales
        np.testing.assert_allclose(sample_means, expected_means, rtol=0.05)

    def test_normality_samples_have_correct_variance(self, test_data):
        """Test that samples have approximately correct variance."""
        cls_bottom_up = BottomUp()
        P, W = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )

        normality = Normality(
            S=test_data["S"],
            P=P,
            W=W,
            y_hat=test_data["y_hat_base"],
            sigmah=test_data["sigmah"],
            covariance_type="diagonal",
            seed=42,
        )

        # Generate many samples for statistical test
        num_samples = 10000
        samples = normality.get_samples(num_samples=num_samples)
        sample_stds = samples.std(axis=2)

        # Expected standard deviation
        expected_stds = normality.sigmah_rec

        # Check that sample stds are close to expected stds
        np.testing.assert_allclose(sample_stds, expected_stds, rtol=0.10)

    def test_normality_correlation_matrix_is_valid(self, test_data):
        """Test that computed correlation matrix has valid properties."""
        cls_bottom_up = BottomUp()
        P, _ = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )
        residuals = test_data["y_base"] - test_data["y_hat_base_insample"]

        for cov_type in ["full", "shrink"]:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                normality = Normality(
                    S=test_data["S"],
                    P=P,
                    y_hat=test_data["y_hat_base"],
                    sigmah=test_data["sigmah"],
                    covariance_type=cov_type,
                    residuals=residuals,
                )

            corr = normality._correlation_matrix

            # Check diagonal is 1
            np.testing.assert_allclose(np.diag(corr), 1.0, atol=1e-10)

            # Check symmetry
            np.testing.assert_allclose(corr, corr.T, atol=1e-10)

            # Check values in [-1, 1]
            assert np.all(corr >= -1.0)
            assert np.all(corr <= 1.0)

    def test_normality_reproducibility_with_seed(self, test_data):
        """Test that setting seed produces reproducible samples."""
        cls_bottom_up = BottomUp()
        P, W = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )

        normality1 = Normality(
            S=test_data["S"],
            P=P,
            W=W,
            y_hat=test_data["y_hat_base"],
            sigmah=test_data["sigmah"],
            seed=42,
        )
        samples1 = normality1.get_samples(num_samples=100)

        normality2 = Normality(
            S=test_data["S"],
            P=P,
            W=W,
            y_hat=test_data["y_hat_base"],
            sigmah=test_data["sigmah"],
            seed=42,
        )
        samples2 = normality2.get_samples(num_samples=100)

        np.testing.assert_array_equal(samples1, samples2)

    def test_normality_different_seeds_different_samples(self, test_data):
        """Test that different seeds produce different samples."""
        cls_bottom_up = BottomUp()
        P, W = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )

        normality1 = Normality(
            S=test_data["S"],
            P=P,
            W=W,
            y_hat=test_data["y_hat_base"],
            sigmah=test_data["sigmah"],
            seed=42,
        )
        samples1 = normality1.get_samples(num_samples=100)

        normality2 = Normality(
            S=test_data["S"],
            P=P,
            W=W,
            y_hat=test_data["y_hat_base"],
            sigmah=test_data["sigmah"],
            seed=123,
        )
        samples2 = normality2.get_samples(num_samples=100)

        assert not np.allclose(samples1, samples2)

    # get_prediction_levels and get_prediction_quantiles tests

    def test_normality_get_prediction_levels(self, test_data):
        """Test get_prediction_levels method."""
        cls_bottom_up = BottomUp()
        P, W = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )

        normality = Normality(
            S=test_data["S"],
            P=P,
            W=W,
            y_hat=test_data["y_hat_base"],
            sigmah=test_data["sigmah"],
        )

        res = {"mean": normality.SP @ test_data["y_hat_base"]}
        levels = [80, 95]
        res = normality.get_prediction_levels(res, levels)

        # Check that required keys are present
        assert "sigmah" in res
        assert "lo-80" in res
        assert "hi-80" in res
        assert "lo-95" in res
        assert "hi-95" in res

        # Check shapes
        n_series, n_horizon = test_data["y_hat_base"].shape
        assert res["sigmah"].shape == (test_data["S"].shape[0], n_horizon)
        assert res["lo-80"].shape == (test_data["S"].shape[0], n_horizon)
        assert res["hi-80"].shape == (test_data["S"].shape[0], n_horizon)

        # Check that intervals are symmetric around mean
        assert np.allclose(res["hi-80"] - res["mean"], res["mean"] - res["lo-80"])

    def test_normality_get_prediction_quantiles(self, test_data):
        """Test get_prediction_quantiles method."""
        cls_bottom_up = BottomUp()
        P, W = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )

        normality = Normality(
            S=test_data["S"],
            P=P,
            W=W,
            y_hat=test_data["y_hat_base"],
            sigmah=test_data["sigmah"],
        )

        res = {"mean": normality.SP @ test_data["y_hat_base"]}
        quantiles = np.array([0.1, 0.5, 0.9])
        res = normality.get_prediction_quantiles(res, quantiles)

        # Check that required keys are present
        assert "sigmah" in res
        assert "quantiles" in res

        # Check shapes
        n_series = test_data["S"].shape[0]
        n_horizon = test_data["y_hat_base"].shape[1]
        assert res["quantiles"].shape == (n_series, n_horizon, len(quantiles))

        # Check that median (0.5) is roughly equal to mean
        np.testing.assert_allclose(res["quantiles"][:, :, 1], res["mean"], rtol=0.01)

    def test_normality_sparse_matrix_handling(self, test_data):
        """Test that sparse S matrix is handled correctly."""
        cls_bottom_up = BottomUp()
        P, W = cls_bottom_up._get_PW_matrices(
            S=test_data["S"]        )

        # Convert S to sparse matrix
        S_sparse = sp.csr_matrix(test_data["S"])

        normality = Normality(
            S=S_sparse,
            P=P,
            W=W,
            y_hat=test_data["y_hat_base"],
            sigmah=test_data["sigmah"],
        )

        samples = normality.get_samples(num_samples=50)
        assert samples.shape == (test_data["S"].shape[0], test_data["h"], 50)


def test_coherent_samples_shape(samplers):
    """Test that coherent samples have the correct shape."""
    normality_samples = samplers["normality_sampler"].get_samples(num_samples=100)
    bootstrap_samples = samplers["bootstrap_sampler"].get_samples(num_samples=100)
    permbu_samples = samplers["permbu_sampler"].get_samples(num_samples=100)

    assert bootstrap_samples.shape == normality_samples.shape
    assert bootstrap_samples.shape == permbu_samples.shape


class TestConformal:
    """Tests for Conformal class."""

    def test_conformal_basic_functionality(self, test_data):
        """Test basic conformal reconciliation functionality."""
        cls_bottom_up = BottomUp()
        P, W = cls_bottom_up._get_PW_matrices(S=test_data["S"])

        conformal = Conformal(
            S=test_data["S"],
            P=P,
            y_hat=test_data["y_hat_base"],
            y_cal=test_data["y_base"],
            y_hat_cal=test_data["y_hat_base_insample"],
        )
        samples = conformal.get_samples(num_samples=50)
        assert samples.shape == (test_data["S"].shape[0], test_data["h"], 50)

    def test_conformal_prediction_levels(self, test_data):
        """Test that prediction levels are computed correctly."""
        cls_bottom_up = BottomUp()
        P, W = cls_bottom_up._get_PW_matrices(S=test_data["S"])

        conformal = Conformal(
            S=test_data["S"],
            P=P,
            y_hat=test_data["y_hat_base"],
            y_cal=test_data["y_base"],
            y_hat_cal=test_data["y_hat_base_insample"],
        )

        res = {"mean": conformal.y_hat_rec}
        res = conformal.get_prediction_levels(res, level=[90, 95])

        assert "lo-90" in res
        assert "hi-90" in res
        assert "lo-95" in res
        assert "hi-95" in res
        assert np.all(res["hi-90"] >= res["lo-90"])
        assert np.all(res["hi-95"] >= res["lo-95"])

    def test_conformal_insufficient_calibration(self, test_data):
        """Test that insufficient calibration observations raises ValueError."""
        cls_bottom_up = BottomUp()
        P, W = cls_bottom_up._get_PW_matrices(S=test_data["S"])

        with pytest.raises(ValueError) as exc_info:
            Conformal(
                S=test_data["S"],
                P=P,
                y_hat=test_data["y_hat_base"],
                y_cal=np.random.randn(7, 1),
                y_hat_cal=np.random.randn(7, 1),
            )
        assert "At least 2 observations are required" in str(exc_info.value)

    def test_conformal_reproducibility_with_seed(self, test_data):
        """Test that setting seed produces reproducible samples."""
        cls_bottom_up = BottomUp()
        P, W = cls_bottom_up._get_PW_matrices(S=test_data["S"])

        conformal1 = Conformal(
            S=test_data["S"],
            P=P,
            y_hat=test_data["y_hat_base"],
            y_cal=test_data["y_base"],
            y_hat_cal=test_data["y_hat_base_insample"],
            seed=42,
        )

        conformal2 = Conformal(
            S=test_data["S"],
            P=P,
            y_hat=test_data["y_hat_base"],
            y_cal=test_data["y_base"],
            y_hat_cal=test_data["y_hat_base_insample"],
            seed=42,
        )

        samples1 = conformal1.get_samples(num_samples=50)
        samples2 = conformal2.get_samples(num_samples=50)
        np.testing.assert_array_equal(samples1, samples2)

    def test_reconciliation_coherency_via_diagnostics(self, strict_hierarchy_data):
        """Test that reconciliation produces coherent forecasts using diagnostics.

        Uses the coherency diagnostics infrastructure to verify that reconciled
        forecasts satisfy the hierarchical constraints: S @ bottom = all.
        This follows the established pattern in test_core.py.
        """
        Y_hat_df = strict_hierarchy_data["Y_hat_df"]
        Y_train_df = strict_hierarchy_data["Y_train_df"]
        S_df = strict_hierarchy_data["S_df"]
        tags = strict_hierarchy_data["tags"]

        hrec = HierarchicalReconciliation(reconcilers=[BottomUp()])
        reconciled = hrec.reconcile(
            Y_hat_df=Y_hat_df,
            Y_df=Y_train_df,
            S_df=S_df,
            tags=tags,
            diagnostics=True,
        )

        # Verify reconciliation completed
        reconciled_nw = nw.from_native(reconciled)
        assert "y_model/BottomUp" in reconciled_nw.columns

        # Verify coherency using diagnostics (consistent with test_core.py pattern)
        diag = nw.from_native(hrec.diagnostics)
        is_coherent = diag.filter(
            (nw.col("metric") == "is_coherent") & (nw.col("level") == "Overall")
        )["y_model/BottomUp"].to_list()[0]
        assert is_coherent == 1.0, "Reconciled forecasts should be coherent"

    def test_conformal_integration_via_intervals_method(self, strict_hierarchy_data):
        """Test that conformal works through the HierarchicalReconciliation API."""
        Y_hat_df = strict_hierarchy_data["Y_hat_df"]
        Y_train_df = strict_hierarchy_data["Y_train_df"]
        S_df = strict_hierarchy_data["S_df"]
        tags = strict_hierarchy_data["tags"]

        hrec = HierarchicalReconciliation(reconcilers=[BottomUp()])
        reconciled = hrec.reconcile(
            Y_hat_df=Y_hat_df,
            Y_df=Y_train_df,
            S_df=S_df,
            tags=tags,
            level=[90],
            intervals_method="conformal",
        )

        reconciled_nw = nw.from_native(reconciled)
        # Check that conformal intervals were computed
        assert "y_model/BottomUp-lo-90" in reconciled_nw.columns
        assert "y_model/BottomUp-hi-90" in reconciled_nw.columns
        # Check that hi >= lo
        lo_vals = reconciled_nw["y_model/BottomUp-lo-90"].to_numpy()
        hi_vals = reconciled_nw["y_model/BottomUp-hi-90"].to_numpy()
        assert np.all(hi_vals >= lo_vals)

    def test_conformal_prediction_quantiles(self, test_data):
        """Test that prediction quantiles are computed correctly."""
        cls_bottom_up = BottomUp()
        P, W = cls_bottom_up._get_PW_matrices(S=test_data["S"])

        conformal = Conformal(
            S=test_data["S"],
            P=P,
            y_hat=test_data["y_hat_base"],
            y_cal=test_data["y_base"],
            y_hat_cal=test_data["y_hat_base_insample"],
        )

        quantiles = np.array([0.025, 0.5, 0.975])
        res = {"mean": conformal.y_hat_rec}
        res = conformal.get_prediction_quantiles(res, quantiles)

        assert "quantiles" in res
        assert res["quantiles"].shape == (
            test_data["S"].shape[0],
            test_data["h"],
            len(quantiles),
        )
        # Check quantiles are monotonically increasing
        for i in range(len(quantiles) - 1):
            assert np.all(res["quantiles"][:, :, i] <= res["quantiles"][:, :, i + 1])

    def test_conformal_invalid_num_samples(self, test_data):
        """Test that invalid num_samples raises ValueError."""
        cls_bottom_up = BottomUp()
        P, W = cls_bottom_up._get_PW_matrices(S=test_data["S"])

        conformal = Conformal(
            S=test_data["S"],
            P=P,
            y_hat=test_data["y_hat_base"],
            y_cal=test_data["y_base"],
            y_hat_cal=test_data["y_hat_base_insample"],
        )

        with pytest.raises(ValueError) as exc_info:
            conformal.get_samples(num_samples=0)
        assert "positive integer" in str(exc_info.value)

        with pytest.raises(ValueError) as exc_info:
            conformal.get_samples(num_samples=-5)
        assert "positive integer" in str(exc_info.value)




__all__ = ['aggregate', 'aggregate_temporal', 'make_future_dataframe', 'get_cross_temporal_tags', 'HierarchicalPlot', 'set_num_threads', 'get_num_threads']


import itertools

# Default OpenMP to physical cores (logical / 2) to avoid oversubscription,
# matching PyTorch's convention. This matters when running reconciliation
# inside multiprocessing.Pool — without it each worker silently spawns
# all-CPU OpenMP threads.
import os as _os
import reprlib
import sys
import timeit
from collections.abc import Sequence

import matplotlib.pyplot as plt
import narwhals.stable.v2 as nw
import numpy as np
import pandas as pd
import utilsforecast.processing as ufp
import utilsforecast.validation as ufv
from narwhals.typing import Frame, FrameT
from scipy import sparse
from sklearn.preprocessing import OneHotEncoder

from hierarchicalforecast._lib import reconciliation as _lib_recon

_default_threads = int(
    _os.environ.get(
        "OMP_NUM_THREADS", max(1, (_os.cpu_count() or 1) // 2)
    )
)
_lib_recon.set_num_threads(_default_threads)
del _os, _default_threads


def set_num_threads(n: int) -> None:
    """Set the number of OpenMP threads used by the C++ backend.

    Args:
        n (int): Number of threads. Must be >= 1.
    """
    _lib_recon.set_num_threads(n)


def get_num_threads() -> int:
    """Return the current maximum number of OpenMP threads.

    Returns:
        int: Number of threads.
    """
    return _lib_recon.get_num_threads()


class CodeTimer:
    def __init__(self, name=None, verbose=True):
        self.name = " '" + name + "'" if name else ""
        self.verbose = verbose

    def __enter__(self):
        self.start = timeit.default_timer()

    def __exit__(self, exc_type, exc_value, traceback):
        self.took = timeit.default_timer() - self.start
        if self.verbose:
            print(
                "Code block"
                + self.name
                + f" took:\t{self.took:.5f}"
                + " seconds"
            )


def _construct_adjacency_matrix(
    S: sparse.csr_matrix, tags: dict[str, np.ndarray]
) -> sparse.csr_matrix:
    """Construct a disaggregation adjacency matrix.

    Args:
        S (sparse.csr_matrix): A summing matrix for hierarchical or grouped time series.
        tags (dict[str, np.ndarray]): A mapping of level name to node indices.

    Returns:
        sparse.csr_matrix: The disaggregation adjacency matrix for the structure.
    """
    # Get the nodes in each level.
    l = sorted(tags.values(), key=lambda x: len(x))
    # Get the number of aggregation nodes.
    n_a = S.shape[0] - S.shape[1]
    # Copy and cast the summing matrix to bool.
    S = S.astype(bool)
    # Precompute the transpose of the boolean summing matrix.
    S_T = S.T
    # Find the affinity, i.e., connectivity, between nodes in successive
    # levels, construct a sparse block diagonal matrix from these blocks, and
    # return the truncated disaggregation adjacency matrix.
    return sparse.hstack(
        (
            sparse.csr_matrix((n_a, len(l[0])), dtype=bool),
            sparse.block_diag(
                [S[l[i]] * S_T[:, l[i + 1]] for i in range(len(l) - 1)], "csr"
            ),
        )
    )


def is_strictly_hierarchical(S: np.ndarray, tags: dict[str, np.ndarray]) -> bool:
    # main idea:
    # if S represents a strictly hierarchical structure
    # the number of paths before the bottom level
    # should be equal to the number of nodes
    # of the previuos level
    levels_ = dict(sorted(tags.items(), key=lambda x: len(x[1])))
    # removing bottom level
    levels_.popitem()
    # making S categorical
    hiers = [np.argmax(S[idx], axis=0) + 1 for _, idx in levels_.items()]
    hiers = np.vstack(hiers)
    paths = np.unique(hiers, axis=1).shape[1]
    nodes = levels_.popitem()[1].size
    return paths == nodes


def _is_strictly_hierarchical(
    A: sparse.csr_matrix, tags: dict[str, np.ndarray]
) -> bool:
    """Check if a disaggregation structure is strictly hierarchical.

    The nodes in a strictly hierarchical disaggregation structure, except for
    the root node(s), should have exactly one incoming edge.

    Args:
        A (sparse.csr_matrix): A disaggregation adjacency matrix.
        tags (dict[str, np.ndarray]): A mapping of level name to node indices.

    Returns:
        bool: `True` if strictly hierarchical, otherwise `False`.
    """
    return np.all(A.sum(axis=0).A1[len(next(iter(tags.values()))) :] == 1)


def _to_upper_hierarchy(
    bottom_split: list[str], bottom_values: str, upper_key: str
) -> list[str]:
    upper_split = upper_key.split("/")
    upper_idxs = [bottom_split.index(i) for i in upper_split]

    def join_upper(bottom_value):
        bottom_parts = bottom_value.split("/")
        return "/".join(bottom_parts[i] for i in upper_idxs)

    return [join_upper(val) for val in bottom_values]


def aggregate(
    df: Frame,
    spec: list[list[str]],
    exog_vars: dict[str, str | list[str]] | None = None,
    sparse_s: bool = False,
    id_col: str = "unique_id",
    time_col: str = "ds",
    id_time_col: str | None = None,
    target_cols: Sequence[str] = ("y",),
) -> tuple[FrameT, FrameT, dict]:
    """Utils Aggregation Function.

    Aggregates bottom level series contained in the DataFrame `df` according
    to levels defined in the `spec` list.

    Args:
        df (Frame): Dataframe with columns `[time_col, *target_cols]`, columns to aggregate and optionally exog_vars.
        spec (list[list[str]]): list of levels. Each element of the list should contain a list of columns of `df` to aggregate.
        exog_vars (Optional[dict[str, Union[str, list[str]]]], optional): dictionary of string keys & values that can either be a list of strings or a single string
            keys correspond to column names and the values represent the aggregation(s) that will be applied to each column. Accepted values are those from Pandas or Polars aggregation Functions, check the respective docs for guidance. Default is None.
        sparse_s (bool, optional): Return `S_df` as a sparse Pandas dataframe. Default is False.
        id_col (str, optional): Column that will identify each serie after aggregation. Default is "unique_id".
        time_col (str, optional): Column that identifies each timestep, its values can be timestamps or integers. Default is "ds".
        id_time_col (Optional[str], optional): Column that will identify each timestep after temporal aggregation. If provided, aggregate will operate temporally. Default is None.
        target_cols (Sequence[str], optional): list of columns that contains the targets to aggregate. Default is ("y",).

    Returns:
        tuple[FrameT, FrameT, dict]: Y_df, S_df, tags
            Y_df: Hierarchically structured series.
            S_df: Summing dataframe.
            tags: Aggregation indices.
    """
    # To Narwhals
    target_cols = list(target_cols)
    df_nw = nw.from_native(df)
    backend = df_nw.implementation

    # Check if we do temporal aggregation
    if id_time_col is not None:
        temporal_agg = True
        group_col = id_col
        _id_col = id_time_col
    else:
        temporal_agg = False
        group_col = time_col
        _id_col = id_col

    # Checks
    # Generate order-preserving list of unique cols based on spec
    spec_cols = list(dict.fromkeys([col for cols in spec for col in cols]))
    # Check if last level in spec contains all levels
    missing_cols_in_bottom_spec = set(spec_cols) - set(spec[-1])
    if missing_cols_in_bottom_spec and not temporal_agg:
        raise ValueError(
            f"Check the last (bottom) level of spec, it has missing columns: {reprlib.repr(missing_cols_in_bottom_spec)}"
        )

    if sparse_s and not nw.dependencies.is_pandas_dataframe(df):
        raise ValueError("Sparse output is only supported for Pandas DataFrames.")

    for col in df_nw.columns:
        if df_nw[col].is_null().any():
            raise ValueError(
                f"Column {col} contains null values. Make sure no column in the DataFrame contains null values."
            )

    # Check whether all columns in the spec are in the df
    aggregation_cols_in_spec = list(
        dict.fromkeys([col for cols in spec for col in cols])
    )
    for col in aggregation_cols_in_spec:
        if col not in df_nw.columns:
            raise ValueError(f"Column {col} in spec not present in df")

    # Prepare the aggregation dictionary
    agg_dict = dict(
        zip(target_cols, tuple(zip(target_cols, len(target_cols) * ["sum"], strict=False)), strict=False)
    )

    # Check if exog_vars are present in df & add to the aggregation dictionary if it is not None
    exog_var_names = []
    if exog_vars is not None:
        missing_vars = [var for var in exog_vars.keys() if var not in df.columns]
        if missing_vars:
            raise ValueError(
                f"The following exogenous variables are not present in the DataFrame: {', '.join(missing_vars)}"
            )
        else:
            # Update agg_dict to handle multiple aggregations for each exog_vars key
            for key, agg_func in exog_vars.items():
                # Ensure agg_func is a list
                if isinstance(
                    agg_func, str
                ):  # If it's a single string, convert to list
                    agg_func = [agg_func]
                elif not isinstance(agg_func, list):  # Raise an error if it's neither
                    raise ValueError(
                        f"Aggregation functions for '{key}' must be a string or a list of strings."
                    )

                for func in agg_func:
                    agg_dict[f"{key}_{func}"] = (
                        key,
                        func,
                    )  # Update the agg_dict with the new naming structure
                    exog_var_names.append(f"{key}_{func}")

    # compute aggregations and tags
    spec = sorted(spec, key=len)

    tags = {}
    Y_nws = []
    category_list = []
    level_sep = "/"
    # Perform the aggregation
    for level in spec:
        level_name = level_sep.join(level)

        # Create Y_df
        Y_level = df_nw.with_columns(
            *[
                getattr(nw.col(col), agg)().over(level + [group_col]).alias(col_name)
                for col_name, (col, agg) in agg_dict.items()
            ]
        )
        Y_level = Y_level.select(
            nw.concat_str([nw.col(col) for col in level], separator=level_sep).alias(
                _id_col
            ),
            nw.all(),
        )
        # For temporal aggregation, we need to keep the time column
        if temporal_agg:
            Y_level = Y_level.sort(by=[id_col, time_col])
            Y_level = Y_level.unique(
                subset=level + [group_col], maintain_order=temporal_agg, keep="last"
            ).select([_id_col, group_col, time_col, *target_cols] + exog_var_names)
            tags[level_name] = (
                Y_level[_id_col].unique(maintain_order=temporal_agg).to_numpy()
            )
        else:
            Y_level = Y_level.unique(
                subset=level + [group_col], maintain_order=temporal_agg, keep="last"
            ).select([_id_col, group_col, *target_cols] + exog_var_names)
            Y_level = Y_level.sort(by=[_id_col, group_col])
            tags[level_name] = (
                Y_level[_id_col].unique(maintain_order=temporal_agg).sort().to_numpy()
            )

        Y_nws.append(Y_level)
        category_list.extend(tags[level_name])

    Y_nw = nw.concat(Y_nws, how="vertical")
    Y_nw = nw.maybe_reset_index(Y_nw)
    Y_df = Y_nw.to_native()

    # construct S
    if temporal_agg:
        bottom = spec_cols
        bottom_levels = (
            df_nw.select(
                nw.concat_str(
                    [nw.col(col) for col in bottom], separator=level_sep
                ).alias(_id_col)
            )
            .unique(subset=[_id_col], maintain_order=temporal_agg, keep="last")[_id_col]
            .to_numpy()
        )
        categories = list([v for k, v in tags.items() if k != level_name])
        categories += [bottom_levels]
    else:
        bottom = spec[-1]
        bottom_levels = tags[level_name]
        categories = list(tags.values())

    S = np.empty((len(bottom_levels), len(spec)), dtype=object)

    for j, levels in enumerate(spec[:-1]):
        S[:, j] = _to_upper_hierarchy(bottom, bottom_levels, level_sep.join(levels))
    S[:, -1] = bottom_levels

    encoder = OneHotEncoder(
        categories=categories, sparse_output=sparse_s, dtype=np.float64
    )
    S_dum = encoder.fit_transform(S)

    if not sparse_s:
        S_nw = nw.from_dict(
            {
                **{_id_col: category_list},
                **dict(zip(tags[level_name], S_dum, strict=False)),
            },
            backend=backend,
        )
        S_nw = nw.maybe_reset_index(S_nw)
        S_df = S_nw.to_native()
    else:
        S_df = pd.DataFrame.sparse.from_spmatrix(
            S_dum.T, columns=list(bottom_levels), index=category_list
        )
        S_df = S_df.reset_index(names=_id_col)

    return Y_df, S_df, tags


def aggregate_temporal(
    df: Frame,
    spec: dict[str, int],
    exog_vars: dict[str, str | list[str]] | None = None,
    sparse_s: bool = False,
    id_col: str = "unique_id",
    time_col: str = "ds",
    id_time_col: str = "temporal_id",
    target_cols: Sequence[str] = ("y",),
    aggregation_type: str = "local",
) -> tuple[FrameT, FrameT, dict]:
    """Utils Aggregation Function for Temporal aggregations.

    Aggregates bottom level timesteps contained in the DataFrame `df` according
    to temporal levels defined in the `spec` list.

    Args:
        df (Frame): Dataframe with columns `[time_col, target_cols]` and columns to aggregate.
        spec (dict[str, int]): Dictionary of temporal levels. Each key should be a string with the value representing the number of bottom-level timesteps contained in the aggregation.
        exog_vars (Optional[dict[str, Union[str, list[str]]]], optional): dictionary of string keys & values that can either be a list of strings or a single string
            keys correspond to column names and the values represent the aggregation(s) that will be applied to each column. Accepted values are those from Pandas or Polars aggregation Functions, check the respective docs for guidance. Default is None.
        sparse_s (bool, optional): Return `S_df` as a sparse Pandas dataframe. Default is False.
        id_col (str, optional): Column that will identify each serie after aggregation. Default is 'unique_id'.
        time_col (str, optional): Column that identifies each timestep, its values can be timestamps or integers. Default is 'ds'.
        id_time_col (str, optional): Column that will identify each timestep after aggregation. Default is 'temporal_id'.
        target_cols (Sequence[str], optional): List of columns that contain the targets to aggregate. Default is ('y',).
        aggregation_type (str, optional): If 'local' the aggregation will be performed on the timestamps of each timeseries independently. If 'global' the aggregation will be performed on the unique timestamps of all timeseries. Default is 'local'.

    Returns:
        tuple[FrameT, FrameT, dict]: Y_df, S_df, tags
            Y_df: Temporally hierarchically structured series.
            S_df: Temporal summing dataframe.
            tags: Temporal aggregation indices.
    """
    # Check if ds column is a timestamp or integer, if not raise an error
    target_cols = list(target_cols)
    df = ufv.ensure_time_dtype(df=df, time_col=time_col)
    df = ufp.ensure_sorted(df=df, id_col=id_col, time_col=time_col)
    df_nw = nw.from_native(df)

    # We add a cumulative count column to the dataframe to be able to compute the aggregations
    if aggregation_type == "local":
        df_nw = df_nw.with_columns(
            nw.col(time_col).cum_count().over([id_col]).alias(f"{time_col}_count")
        )
    elif aggregation_type == "global":
        unique_ts = df_nw.select(nw.col(time_col)).unique(maintain_order=True)
        unique_ts = unique_ts.with_columns(
            nw.col(time_col).cum_count().alias(f"{time_col}_count")
        )
        df_nw = df_nw.join(unique_ts, on=[time_col], how="left")
        df_nw = df_nw.sort(by=[id_col, time_col])
    else:
        raise ValueError("aggregation_type must be either 'local' or 'global'.")
    df_nw = nw.maybe_reset_index(df_nw)

    # Check spec that lowest level with seasonality of 1 has been defined
    if 1 not in spec.values():
        raise ValueError(
            "The spec must contain a level with a seasonality of 1. This represents the lowest level in the temporal aggregation."
        )

    # Check for duplicate values in spec
    if len(spec) != len(set(spec.values())):
        raise ValueError(
            "The spec must contain unique values for each level. Each value represents the seasonality of the aggregation."
        )

    # Check for duplicate keys in spec
    if len(spec) != len(set(spec.keys())):
        raise ValueError(
            "The spec must contain unique keys for each level. Each key represents the name of the aggregation."
        )

    # Loop over the spec and create the aggregation columns
    spec_agg: list = []
    for agg, seasonality in spec.items():
        df_nw = df_nw.with_columns(
            nw.concat_str(
                [
                    nw.lit(agg),
                    nw.lit("-"),
                    (((nw.col(f"{time_col}_count") - 1) // seasonality) + 1).cast(
                        nw.String
                    ),
                ]
            ).alias(agg)
        )
        spec_agg.append([agg])

    # If target_cols is not in df, we add a placeholder column so that we can compute the aggregations
    add_placeholder = False
    if set(df.columns) == set([time_col, id_col]):
        add_placeholder = True
        df_nw = df_nw.with_columns(nw.lit(0).alias("y"))

    # Create the aggregation
    df = df_nw.to_native()
    Y_df, S_df, tags = aggregate(
        df=df,
        spec=spec_agg,
        exog_vars=exog_vars,
        sparse_s=sparse_s,
        id_col=id_col,
        time_col=time_col,
        target_cols=target_cols,
        id_time_col=id_time_col,
    )
    Y_nw = nw.from_native(Y_df)

    # Drop the placeholder column if it was added
    if add_placeholder:
        Y_nw = Y_nw.drop("y")

    Y_df = Y_nw.to_native()

    return Y_df, S_df, tags


def make_future_dataframe(
    df: Frame,
    freq: str | int,
    h: int,
    id_col: str = "unique_id",
    time_col: str = "ds",
) -> FrameT:
    """Create future dataframe for forecasting.

    Args:
        df (Frame): Dataframe with ids, times and values for the exogenous regressors.
        freq (Union[str, int]): Frequency of the data. Must be a valid pandas or polars offset alias, or an integer.
        h (int): Forecast horizon.
        id_col (str, optional): Column that identifies each serie. Default is 'unique_id'.
        time_col (str, optional): Column that identifies each timestep, its values can be timestamps or integers. Default is 'ds'.

    Returns:
        FrameT: DataFrame with future values.
    """
    times_by_id = ufp.group_by_agg(df, id_col, {time_col: "max"}, maintain_order=True)
    times_by_id = ufp.sort(times_by_id, id_col)
    future_df = ufp.make_future_dataframe(
        uids=times_by_id[id_col],
        last_times=times_by_id[time_col],
        freq=freq,
        h=h,
        id_col=id_col,
        time_col=time_col,
    )
    return future_df


def get_cross_temporal_tags(
    df: Frame,
    tags_cs: dict[str, np.ndarray],
    tags_te: dict[str, np.ndarray],
    sep: str = "//",
    id_col: str = "unique_id",
    id_time_col: str = "temporal_id",
    cross_temporal_id_col: str = "cross_temporal_id",
) -> tuple[FrameT, dict[str, np.ndarray]]:
    """Get cross-temporal tags.

    Args:
        df (Frame): DataFrame with temporal ids.
        tags_cs (dict[str, np.ndarray]): Tags for the cross-sectional hierarchies.
        tags_te (dict[str, np.ndarray]): Tags for the temporal hierarchies.
        sep (str, optional): Separator for the cross-temporal tags. Default is "//".
        id_col (str, optional): Column that identifies each series. Default is 'unique_id'.
        id_time_col (str, optional): Column that identifies each (aggregated) timestep. Default is 'temporal_id'.
        cross_temporal_id_col (str, optional): Column that will identify each cross-temporal aggregation. Default is 'cross_temporal_id'.

    Returns:
        tuple[FrameT, dict[str, np.ndarray]]: df, tags_ct
            df: DataFrame with cross-temporal ids.
            tags_ct: Tags for the cross-temporal hierarchies.
    """
    df_nw = nw.from_native(df)

    # Check if relevant columns are present
    if id_col not in df_nw.columns:
        raise ValueError(f"Column '{id_col}' not present in df")
    if id_time_col not in df_nw.columns:
        raise ValueError(f"Column '{id_time_col}' not present in df")

    # Create cross-temporal tags
    tags_ct = {}
    for key_cs, value_cs in tags_cs.items():
        for key_te, value_te in tags_te.items():
            key_ct = key_cs + sep + key_te
            value_ct = list(sep.join(s) for s in itertools.product(value_cs, value_te))
            tags_ct[key_ct] = value_ct

    df_nw = df_nw.with_columns(
        **{cross_temporal_id_col: df_nw[id_col] + sep + df_nw[id_time_col]}
    )
    df = df_nw.to_native()

    return df, tags_ct


class HierarchicalPlot:
    r"""Hierarchical Plot

    This class contains a collection of matplotlib visualization methods, suited for small
    to medium sized hierarchical series.

    Args:
        S (Frame): DataFrame with summing matrix of size `(base, bottom)`, see [aggregate function](./utils#function-aggregate).
        tags (dict[str, np.ndarray]): hierarchical aggregation indexes, where
            each key is a level and its value contains tags associated to that level.
        S_id_col (str, optional): column that identifies each aggregation. Default is 'unique_id'.
    """

    def __init__(
        self,
        S: Frame,
        tags: dict[str, np.ndarray],
        S_id_col: str = "unique_id",
    ):

        self.S = nw.from_native(S)
        S_cols_ex_id_col = self.S.columns
        S_cols_ex_id_col.remove(S_id_col)
        self.S_cols_ex_id_col = S_cols_ex_id_col
        self.tags = tags

    def plot_summing_matrix(self):
        r"""Summation Constraints plot

        This method simply plots the hierarchical aggregation
        constraints matrix $\mathbf{S}$.

        Returns:
            fig (matplotlib.figure.Figure): figure object containing the plot of the summing matrix.
        """
        fig = plt.figure(num=1, figsize=(4, 6), dpi=80, facecolor="w")
        plt.spy(self.S[self.S_cols_ex_id_col].to_numpy())
        plt.show()
        return fig

    def plot_series(
        self,
        series: str,
        Y_df: Frame,
        models: list[str] | None = None,
        level: list[int] | None = None,
        id_col: str = "unique_id",
        time_col: str = "ds",
        target_col: str = "y",
    ):
        r"""Single Series plot

        Args:
            series (str): string identifying the `'unique_id'` any-level series to plot.
            Y_df (Frame): hierarchically structured series ($\mathbf{y}_{[a,b]}$).
                    It contains columns `['unique_id', 'ds', 'y']`, it may have `'models'`.
            models (Optional[list[str]], optional): string identifying filtering model columns. Default is None.
            level (Optional[list[int]], optional): confidence levels for prediction intervals available in `Y_df`. Default is None.
            id_col (str, optional): column that identifies each series. Default is 'unique_id'.
            time_col (str, optional): column that identifies each timestep, its values can be timestamps or integers. Default is 'ds'.
            target_col (str, optional): column that contains the target. Default is 'y'.

        Returns:
            fig (matplotlib.figure.Figure): figure object containing the plot of the single series.
        """
        Y_nw = nw.from_native(Y_df)

        if series not in self.S[id_col]:
            raise Exception(f"time series {series} not found")
        fig, ax = plt.subplots(1, 1, figsize=(20, 7))
        df_plot = Y_nw.filter(nw.col(id_col) == series)
        cols = (
            models
            if models is not None
            else [col for col in df_plot.columns if col not in [id_col, time_col]]
        )
        cols_wo_levels = [
            col for col in cols if ("-lo-" not in col and "-hi-" not in col)
        ]
        try:
            cmap = plt.get_cmap("tab10", 10)
        except AttributeError:
            cmap = plt.cm.get_cmap("tab10", 10)
        cmap = [cmap(i) for i in range(10)][: len(cols_wo_levels)]
        cmap_dict = dict(zip(cols_wo_levels, cmap, strict=False))
        for col in cols_wo_levels:
            ax.plot(
                df_plot[time_col].to_numpy(),
                df_plot[col].to_numpy(),
                linewidth=2,
                label=col,
                color=cmap_dict[col],
            )
            if level is not None and col != target_col:
                for lv in level:
                    if f"{col}-lo-{lv}" not in df_plot.columns:
                        # if model
                        # doesnt have levels
                        continue
                    ax.fill_between(
                        df_plot.select(nw.col(time_col))[time_col].to_numpy(),
                        df_plot.select(nw.col(f"{col}-lo-{lv}"))[
                            f"{col}-lo-{lv}"
                        ].to_numpy(),
                        df_plot.select(nw.col(f"{col}-hi-{lv}"))[
                            f"{col}-hi-{lv}"
                        ].to_numpy(),
                        alpha=-lv / 100 + 1,
                        color=cmap_dict[col],
                        label=f"{col}_level_{lv}",
                    )
        ax.set_title(f"{series} Forecast", fontsize=22)
        ax.set_xlabel("Timestamp [t]", fontsize=20)
        ax.legend(prop={"size": 15})
        ax.grid()
        ax.xaxis.set_major_locator(plt.MaxNLocator(min(max(len(df_plot) // 10, 1), 10)))
        for label in ax.get_xticklabels() + ax.get_yticklabels():
            label.set_fontsize(20)

        plt.show()
        return fig

    def plot_hierarchically_linked_series(
        self,
        bottom_series: str,
        Y_df: Frame,
        models: list[str] | None = None,
        level: list[int] | None = None,
        id_col: str = "unique_id",
        time_col: str = "ds",
        target_col: str = "y",
    ):
        r"""Hierarchically Linked Series plot

        Args:
            bottom_series (str): string identifying the `'unique_id'` bottom-level series to plot.
            Y_df (Frame): hierarchically structured series ($\mathbf{y}_{[a,b]}$).
                    It contains columns ['unique_id', 'ds', 'y'] and models.
            models (Optional[list[str]], optional): string identifying filtering model columns. Default is None.
            level (Optional[list[int]], optional): confidence levels for prediction intervals available in `Y_df`. Default is None.
            id_col (str, optional): column that identifies each series. Default is 'unique_id'.
            time_col (str, optional): column that identifies each timestep, its values can be timestamps or integers. Default is 'ds'.
            target_col (str, optional): column that contains the target. Default is 'y'.

        Returns:
            fig (matplotlib.figure.Figure): figure object containing the plots of the hierarchically linked series.
        """
        Y_nw = nw.from_native(Y_df)

        if bottom_series not in self.S.columns:
            raise Exception(f"bottom time series {bottom_series} not found")

        linked_series = (
            self.S[[id_col, bottom_series]]
            .filter(nw.col(bottom_series) == 1)[id_col]
            .to_numpy()
        )
        fig, axs = plt.subplots(
            len(linked_series), 1, figsize=(20, 2 * len(linked_series))
        )
        cols = (
            models
            if models is not None
            else [col for col in Y_nw.columns if col not in [id_col, time_col]]
        )
        cols_wo_levels = [
            col for col in cols if ("-lo-" not in col and "-hi-" not in col)
        ]
        cmap = plt.cm.get_cmap("tab10", 10)
        cmap = [cmap(i) for i in range(10)][: len(cols_wo_levels)]
        cmap_dict = dict(zip(cols_wo_levels, cmap, strict=False))
        for idx, series in enumerate(linked_series):
            df_plot = Y_nw.filter(nw.col(id_col) == series)
            for col in cols_wo_levels:
                axs[idx].plot(
                    df_plot[time_col].to_numpy(),
                    df_plot[col].to_numpy(),
                    linewidth=2,
                    label=col,
                    color=cmap_dict[col],
                )
                if level is not None and col != target_col:
                    for lv in level:
                        if f"{col}-lo-{lv}" not in df_plot.columns:
                            # if model
                            # doesnt have levels
                            continue
                        axs[idx].fill_between(
                            df_plot.select(nw.col(time_col))[time_col].to_numpy(),
                            df_plot.select(nw.col(f"{col}-lo-{lv}"))[
                                f"{col}-lo-{lv}"
                            ].to_numpy(),
                            df_plot.select(nw.col(f"{col}-hi-{lv}"))[
                                f"{col}-hi-{lv}"
                            ].to_numpy(),
                            alpha=-lv / 100 + 1,
                            color=cmap_dict[col],
                            label=f"{col}_level_{lv}",
                        )
            axs[idx].set_title(f"{series}", fontsize=10)
            axs[idx].grid()
            axs[idx].get_xaxis().label.set_visible(False)
            axs[idx].legend().set_visible(False)
            axs[idx].xaxis.set_major_locator(
                plt.MaxNLocator(min(max(len(df_plot) // 10, 1), 10))
            )
            for label in axs[idx].get_xticklabels() + axs[idx].get_yticklabels():
                label.set_fontsize(10)
        plt.subplots_adjust(hspace=0.4)
        handles, labels = axs[0].get_legend_handles_labels()
        kwargs = dict(
            loc="lower center", prop={"size": 10}, bbox_to_anchor=(0, 0.05, 1, 1)
        )
        if sys.version_info.minor > 7:  # noqa: YTT204
            kwargs["ncols"] = np.max([2, np.ceil(len(labels) / 2)])
        fig.legend(handles, labels, **kwargs)
        return fig

    def plot_hierarchical_predictions_gap(
        self,
        Y_df: Frame,
        models: list[str] | None = None,
        xlabel: str | None = None,
        ylabel: str | None = None,
        id_col: str = "unique_id",
        time_col: str = "ds",
        target_col: str = "y",
    ):
        r"""Hierarchically Predictions Gap plot

        Args:
            Y_df (Frame): hierarchically structured series ($\mathbf{y}_{[a,b]}$).
                    It contains columns ['unique_id', 'ds', 'y'] and models.
            models (Optional[list[str]], optional): string identifying filtering model columns. Default is None.
            xlabel (Optional[str], optional): string for the plot's x axis label. Default is None.
            ylabel (Optional[str], optional): string for the plot's y axis label. Default is None.
            id_col (str, optional): column that identifies each series. Default is 'unique_id'.
            time_col (str, optional): column that identifies each timestep, its values can be timestamps or integers. Default is 'ds'.
            target_col (str, optional): column that contains the target. Default is 'y'.

        Returns:
            fig (matplotlib.figure.Figure): figure object containing the plot of the aggregated predictions at different levels of the hierarchical structure.
        """
        Y_nw = nw.from_native(Y_df)

        # Parse predictions dataframe
        horizon_dates = Y_nw["ds"].unique().to_numpy()
        cols = (
            models
            if models is not None
            else [col for col in Y_nw.columns if col not in [id_col, time_col]]
        )

        # Plot predictions across tag levels
        fig, ax = plt.subplots(figsize=(8, 5))

        if target_col in Y_nw.columns:
            idx_top = (
                self.S.with_columns(sum_cols=nw.sum_horizontal(cols))
                .sort(by="sum_cols", descending=True)[0][id_col]
                .to_numpy()
            )
            y_plot = Y_nw.filter(nw.col(id_col) == idx_top)[target_col].to_numpy()
            plt.plot(horizon_dates, y_plot, label="True")

        ys = []
        for tag in self.tags:
            y_plot = sum(
                [
                    Y_nw.filter(nw.col(id_col) == idx)[cols].to_numpy()
                    for idx in self.tags[tag]
                ]
            )
            plt.plot(horizon_dates, y_plot, label=f"Level: {tag}")

            ys.append(y_plot[:, None])

        plt.title("Predictions Accumulated Difference")
        if ylabel is not None:
            plt.ylabel(ylabel)
        if xlabel is not None:
            plt.xlabel(xlabel)

        plt.legend()
        plt.grid()
        plt.show()
        return fig


# convert levels to output quantile names
def level_to_outputs(level: list[int]) -> tuple[list[float], list[str]]:
    """Converts list of levels into output names matching StatsForecast and NeuralForecast methods.

    Args:
        level (list[int]): Probability levels for prediction intervals [0,100].

    Returns:
        tuple[list[float], list[str]]: quantiles and output_names
            quantiles: quantiles derived from levels.
            output_names: String list with output column names.
    """
    qs = sum([[50 - l / 2, 50 + l / 2] for l in level], [])
    output_names = sum([[f"-lo-{l}", f"-hi-{l}"] for l in level], [])

    sort_idx = np.argsort(qs)
    quantiles = np.array(qs)[sort_idx]

    # Add default median
    quantiles = np.concatenate([np.array([50]), quantiles]) / 100
    output_names = list(np.array(output_names)[sort_idx])
    output_names.insert(0, "-median")

    return quantiles, output_names


# convert quantiles to output quantile names
def quantiles_to_outputs(quantiles: list[float]) -> tuple[list[float], list[str]]:
    """Converts list of quantiles into output names matching StatsForecast and NeuralForecast methods.

    Args:
        quantiles (list[float]): Alternative to level, quantiles to estimate from y distribution [0., 1.].

    Returns:
        tuple[list[float], list[str]]: quantiles and output_names
            quantiles: quantiles to estimate from y distribution.
            output_names: String list with output column names.
    """
    output_names = []
    for q in quantiles:
        if q < 0.50:
            output_names.append(f"-lo-{np.round(100-200*q,2)}")
        elif q > 0.50:
            output_names.append(f"-hi-{np.round(100-200*(1-q),2)}")
        else:
            output_names.append("-median")
    return quantiles, output_names


# given input array of sample forecasts and inptut quantiles/levels,
# output a Pandas Dataframe with columns of quantile predictions
def samples_to_quantiles_df(
    samples: np.ndarray,
    unique_ids: Sequence[str],
    dates: list[str],
    quantiles: list[float] | None = None,
    level: list[int] | None = None,
    model_name: str = "model",
    id_col: str = "unique_id",
    time_col: str = "ds",
    backend: str = "pandas",
) -> tuple[list[float], FrameT]:
    """Transform Random Samples into HierarchicalForecast input.

    Auxiliary function to create compatible HierarchicalForecast input `Y_hat_df` dataframe.

    Args:
        samples (np.ndarray): Samples from forecast distribution of shape [n_series, n_samples, horizon].
        unique_ids (Sequence[str]): Unique identifiers for each time series.
        dates (list[str]): list of forecast dates.
        quantiles (Optional[list[float]], optional): Alternative to level, quantiles to estimate from y distribution [0., 1.]. Default is None.
        level (Optional[list[int]], optional): Probability levels for prediction intervals [0,100]. Default is None.
        model_name (str, optional): Name of forecasting model. Default is "model".
        id_col (str, optional): column that identifies each serie. Default is 'unique_id'.
        time_col (str, optional): column that identifies each timestep, its values can be timestamps or integers. Default is 'ds'.
        backend (str, optional): backend to use for the output dataframe, either 'pandas' or 'polars'. Default is 'pandas'.

    Returns:
        tuple[list[float], FrameT]: quantiles and Y_hat_df
            quantiles: quantiles to estimate from y distribution [0., 1.].
            Y_hat_df: DataFrame with base quantile forecasts with columns ds and models to reconcile indexed by unique_id.
    """

    # Get the shape of the array
    n_series, n_samples, horizon = samples.shape

    if n_series != len(unique_ids):
        raise ValueError(
            f"Number of unique_ids ({len(unique_ids)}) must match the number of series ({n_series})."
        )
    if horizon != len(dates):
        raise ValueError(
            f"Number of dates ({len(dates)}) must match third dimension of samples array ({horizon})."
        )
    if not ((quantiles is None) ^ (level is None)):
        raise ValueError("Either quantiles or level must be provided, but not both.")

    namespace = sys.modules.get(backend, None)
    if namespace is None:
        raise ValueError(f"DataFrame backend {backend} not installed.")

    # create initial dictionary
    forecasts_mean = np.mean(samples, axis=1).flatten()
    unique_ids = np.repeat(unique_ids, horizon)
    ds = np.tile(dates, n_series)

    # create quantiles and quantile names
    if level is not None:
        _quantiles, quantile_names = level_to_outputs(level)
    elif quantiles is not None:
        _quantiles, quantile_names = quantiles_to_outputs(quantiles)

    percentiles = [quantile * 100 for quantile in _quantiles]
    col_names = np.array(
        [model_name + quantile_name for quantile_name in quantile_names]
    )

    # add quantiles to dataframe
    forecasts_quantiles = np.percentile(samples, percentiles, axis=1)

    forecasts_quantiles = np.transpose(
        forecasts_quantiles, (1, 2, 0)
    )  # [Q,H,N] -> [N,H,Q]
    forecasts_quantiles = forecasts_quantiles.reshape(-1, len(_quantiles))

    df_nw = nw.from_dict(
        {
            **{id_col: unique_ids, time_col: ds, model_name: forecasts_mean},
            **dict(zip(col_names, forecasts_quantiles.T, strict=False)),
        },
        backend=backend,
    )

    return _quantiles, df_nw.to_native()


# Masked empirical covariance matrix
def _ma_cov(residuals: np.ndarray, not_nan_mask: np.ndarray):
    """Masked empirical covariance matrix.

    :meta private:
    """
    return _lib_recon._ma_cov(residuals, not_nan_mask)


# Shrunk covariance matrix using the Schafer-Strimmer method
def _shrunk_covariance_schaferstrimmer_no_nans(
    residuals: np.ndarray, mint_shr_ridge: float
):
    """Shrink empirical covariance (Schäfer-Strimmer, no NaN handling).

    :meta private:
    """
    return _lib_recon._shrunk_covariance_schaferstrimmer_no_nans(
        residuals, mint_shr_ridge
    )


def _shrunk_covariance_schaferstrimmer_with_nans(
    residuals: np.ndarray, not_nan_mask: np.ndarray, mint_shr_ridge: float
):
    """Shrink empirical covariance (Schäfer-Strimmer, with NaN masking).

    :meta private:
    """
    return _lib_recon._shrunk_covariance_schaferstrimmer_with_nans(
        residuals, not_nan_mask, mint_shr_ridge
    )


# Lasso cyclic coordinate descent
def _lasso(
    X: np.ndarray,
    y: np.ndarray,
    lambda_reg: float,
    max_iters: int = 1_000,
    tol: float = 1e-4,
):
    """Lasso cyclic coordinate descent.

    :meta private:
    """
    return _lib_recon._lasso(np.asfortranarray(X), y, lambda_reg, max_iters, tol)
